"""Drolta SQLite Query Engine.

Drolta is a query engine that wraps a  SQLite database to allow
users to write queries that are more declarative, composable,
and legible than raw SQL. Drolta pulls inspiration
from logic languages like Prolog and Datalog to give users an
easy-to-use database querying experience.

"""

__version__ = "0.4.0"

from .engine import QueryEngine

__all__ = ["QueryEngine"]



"""Drolta Data Classes."""

from __future__ import annotations

from typing import Optional

import attrs

from drolta.ast import (
    ExpressionNode,
    GroupByExpression,
    LimitExpression,
    OrderByExpression,
)


@attrs.define(slots=True)
class ResultVariable:
    """A query rule result variable."""

    var_name: str
    aggregate_name: str = ""
    alias: str = ""

    def __str__(self):
        final_str = self.var_name

        if self.aggregate_name:
            final_str = f"{self.aggregate_name}({final_str})"

        if self.alias:
            final_str = f'{final_str} AS "{self.alias}"'

        return final_str


@attrs.define(slots=True)
class RuleData:
    """A Drolta query rule."""

    name: str
    result_vars: list[ResultVariable]
    where_expressions: list[ExpressionNode]
    order_by: Optional[OrderByExpression]
    group_by: Optional[GroupByExpression]
    limit: Optional[LimitExpression]


@attrs.define(slots=True)
class EngineData:
    """Holds all the data managed by the engine."""

    aliases: dict[str, str] = attrs.field(factory=dict)
    rules: dict[str, RuleData] = attrs.field(factory=dict)



# Generated from ./Drolta.g4 by ANTLR 4.13.2
from antlr4 import *

from drolta.parsing.DroltaParser import DroltaParser


# This class defines a complete listener for a parse tree produced by DroltaParser.
class DroltaListener(ParseTreeListener):

    # Enter a parse tree produced by DroltaParser#prog.
    def enterProg(self, ctx: DroltaParser.ProgContext):
        pass

    # Exit a parse tree produced by DroltaParser#prog.
    def exitProg(self, ctx: DroltaParser.ProgContext):
        pass

    # Enter a parse tree produced by DroltaParser#prog_statement.
    def enterProg_statement(self, ctx: DroltaParser.Prog_statementContext):
        pass

    # Exit a parse tree produced by DroltaParser#prog_statement.
    def exitProg_statement(self, ctx: DroltaParser.Prog_statementContext):
        pass

    # Enter a parse tree produced by DroltaParser#alias_declaration.
    def enterAlias_declaration(self, ctx: DroltaParser.Alias_declarationContext):
        pass

    # Exit a parse tree produced by DroltaParser#alias_declaration.
    def exitAlias_declaration(self, ctx: DroltaParser.Alias_declarationContext):
        pass

    # Enter a parse tree produced by DroltaParser#rule_declaration.
    def enterRule_declaration(self, ctx: DroltaParser.Rule_declarationContext):
        pass

    # Exit a parse tree produced by DroltaParser#rule_declaration.
    def exitRule_declaration(self, ctx: DroltaParser.Rule_declarationContext):
        pass

    # Enter a parse tree produced by DroltaParser#define_clause.
    def enterDefine_clause(self, ctx: DroltaParser.Define_clauseContext):
        pass

    # Exit a parse tree produced by DroltaParser#define_clause.
    def exitDefine_clause(self, ctx: DroltaParser.Define_clauseContext):
        pass

    # Enter a parse tree produced by DroltaParser#query.
    def enterQuery(self, ctx: DroltaParser.QueryContext):
        pass

    # Exit a parse tree produced by DroltaParser#query.
    def exitQuery(self, ctx: DroltaParser.QueryContext):
        pass

    # Enter a parse tree produced by DroltaParser#find_clause.
    def enterFind_clause(self, ctx: DroltaParser.Find_clauseContext):
        pass

    # Exit a parse tree produced by DroltaParser#find_clause.
    def exitFind_clause(self, ctx: DroltaParser.Find_clauseContext):
        pass

    # Enter a parse tree produced by DroltaParser#result_var.
    def enterResult_var(self, ctx: DroltaParser.Result_varContext):
        pass

    # Exit a parse tree produced by DroltaParser#result_var.
    def exitResult_var(self, ctx: DroltaParser.Result_varContext):
        pass

    # Enter a parse tree produced by DroltaParser#where_clause.
    def enterWhere_clause(self, ctx: DroltaParser.Where_clauseContext):
        pass

    # Exit a parse tree produced by DroltaParser#where_clause.
    def exitWhere_clause(self, ctx: DroltaParser.Where_clauseContext):
        pass

    # Enter a parse tree produced by DroltaParser#order_by_statement.
    def enterOrder_by_statement(self, ctx: DroltaParser.Order_by_statementContext):
        pass

    # Exit a parse tree produced by DroltaParser#order_by_statement.
    def exitOrder_by_statement(self, ctx: DroltaParser.Order_by_statementContext):
        pass

    # Enter a parse tree produced by DroltaParser#ordering_term.
    def enterOrdering_term(self, ctx: DroltaParser.Ordering_termContext):
        pass

    # Exit a parse tree produced by DroltaParser#ordering_term.
    def exitOrdering_term(self, ctx: DroltaParser.Ordering_termContext):
        pass

    # Enter a parse tree produced by DroltaParser#group_by_statement.
    def enterGroup_by_statement(self, ctx: DroltaParser.Group_by_statementContext):
        pass

    # Exit a parse tree produced by DroltaParser#group_by_statement.
    def exitGroup_by_statement(self, ctx: DroltaParser.Group_by_statementContext):
        pass

    # Enter a parse tree produced by DroltaParser#limit_statement.
    def enterLimit_statement(self, ctx: DroltaParser.Limit_statementContext):
        pass

    # Exit a parse tree produced by DroltaParser#limit_statement.
    def exitLimit_statement(self, ctx: DroltaParser.Limit_statementContext):
        pass

    # Enter a parse tree produced by DroltaParser#where_statement.
    def enterWhere_statement(self, ctx: DroltaParser.Where_statementContext):
        pass

    # Exit a parse tree produced by DroltaParser#where_statement.
    def exitWhere_statement(self, ctx: DroltaParser.Where_statementContext):
        pass

    # Enter a parse tree produced by DroltaParser#ComparisonFilter.
    def enterComparisonFilter(self, ctx: DroltaParser.ComparisonFilterContext):
        pass

    # Exit a parse tree produced by DroltaParser#ComparisonFilter.
    def exitComparisonFilter(self, ctx: DroltaParser.ComparisonFilterContext):
        pass

    # Enter a parse tree produced by DroltaParser#InFilter.
    def enterInFilter(self, ctx: DroltaParser.InFilterContext):
        pass

    # Exit a parse tree produced by DroltaParser#InFilter.
    def exitInFilter(self, ctx: DroltaParser.InFilterContext):
        pass

    # Enter a parse tree produced by DroltaParser#AndFilter.
    def enterAndFilter(self, ctx: DroltaParser.AndFilterContext):
        pass

    # Exit a parse tree produced by DroltaParser#AndFilter.
    def exitAndFilter(self, ctx: DroltaParser.AndFilterContext):
        pass

    # Enter a parse tree produced by DroltaParser#OrFilter.
    def enterOrFilter(self, ctx: DroltaParser.OrFilterContext):
        pass

    # Exit a parse tree produced by DroltaParser#OrFilter.
    def exitOrFilter(self, ctx: DroltaParser.OrFilterContext):
        pass

    # Enter a parse tree produced by DroltaParser#NotFilter.
    def enterNotFilter(self, ctx: DroltaParser.NotFilterContext):
        pass

    # Exit a parse tree produced by DroltaParser#NotFilter.
    def exitNotFilter(self, ctx: DroltaParser.NotFilterContext):
        pass

    # Enter a parse tree produced by DroltaParser#atom_list.
    def enterAtom_list(self, ctx: DroltaParser.Atom_listContext):
        pass

    # Exit a parse tree produced by DroltaParser#atom_list.
    def exitAtom_list(self, ctx: DroltaParser.Atom_listContext):
        pass

    # Enter a parse tree produced by DroltaParser#comparison_operator.
    def enterComparison_operator(self, ctx: DroltaParser.Comparison_operatorContext):
        pass

    # Exit a parse tree produced by DroltaParser#comparison_operator.
    def exitComparison_operator(self, ctx: DroltaParser.Comparison_operatorContext):
        pass

    # Enter a parse tree produced by DroltaParser#Predicate.
    def enterPredicate(self, ctx: DroltaParser.PredicateContext):
        pass

    # Exit a parse tree produced by DroltaParser#Predicate.
    def exitPredicate(self, ctx: DroltaParser.PredicateContext):
        pass

    # Enter a parse tree produced by DroltaParser#PredicateNot.
    def enterPredicateNot(self, ctx: DroltaParser.PredicateNotContext):
        pass

    # Exit a parse tree produced by DroltaParser#PredicateNot.
    def exitPredicateNot(self, ctx: DroltaParser.PredicateNotContext):
        pass

    # Enter a parse tree produced by DroltaParser#predicate_param.
    def enterPredicate_param(self, ctx: DroltaParser.Predicate_paramContext):
        pass

    # Exit a parse tree produced by DroltaParser#predicate_param.
    def exitPredicate_param(self, ctx: DroltaParser.Predicate_paramContext):
        pass

    # Enter a parse tree produced by DroltaParser#atom.
    def enterAtom(self, ctx: DroltaParser.AtomContext):
        pass

    # Exit a parse tree produced by DroltaParser#atom.
    def exitAtom(self, ctx: DroltaParser.AtomContext):
        pass

    # Enter a parse tree produced by DroltaParser#variable.
    def enterVariable(self, ctx: DroltaParser.VariableContext):
        pass

    # Exit a parse tree produced by DroltaParser#variable.
    def exitVariable(self, ctx: DroltaParser.VariableContext):
        pass


del DroltaParser



# pylint: disable=C0302
"""Unit Tests of Drolta.

This test files does not use pytest fixtures because the database connection
needs to be closed at the end of each test. There is probably a more elegant
way to handle this (probably with a context manager). However, this works for
now.

"""

import sqlite3

import pytest

import drolta
import drolta.engine
from drolta.interpreter import has_alias_cycle


def initialize_test_data(db: sqlite3.Connection) -> None:
    """Initializes the provided data"""

    cursor = db.cursor()

    cursor.executescript(
        """
        DROP TABLE IF EXISTS characters;
        DROP TABLE IF EXISTS houses;
        DROP TABLE IF EXISTS relations;

        CREATE TABLE characters (
            id INTEGER PRIMARY KEY NOT NULL,
            name TEXT,
            house_id INTEGER,
            sex TEXT,
            life_stage TEXT,
            is_alive INTEGER,
            FOREIGN KEY (house_id) REFERENCES houses(id)
        ) STRICT;

        CREATE TABLE houses (
            id INTEGER NOT NULL PRIMARY KEY,
            name TEXT NOT NULL,
            reputation INT NOT NULL,
            is_noble INT NOT NULL
        ) STRICT;

        CREATE TABLE relations (
            from_id INTEGER NOT NULL,
            to_id INTEGER NOT NULL,
            type TEXT NOT NULL,
            FOREIGN KEY (from_id) REFERENCES characters(id),
            FOREIGN KEY (to_id) REFERENCES characters(id)
        ) STRICT;
        """
    )

    cursor.executemany(
        """
        INSERT INTO
        characters (id, name, house_id, sex, life_stage, is_alive)
        VALUES
        (?, ?, ?, ?, ?, ?);
        """,
        [
            (1, "Rhaenyra", 1, "F", "Adult", 1),
            (2, "Laenor", 2, "M", "Adult", 1),
            (3, "Harwin", 3, "M", "Adult", 1),
            (4, "Jacaerys", 2, "M", "Teen", 1),
            (5, "Addam", None, "M", "Teen", 1),
            (6, "Corlys", 2, "M", "Adult", 1),
            (7, "Marilda", None, "F", "Adult", 0),
            (8, "Alyn", None, "M", "Adult", 1),
            (9, "Rhaenys", 1, "F", "Adult", 0),
            (10, "Laena", 2, "F", "Adult", 0),
            (11, "Daemon", 1, "M", "Adult", 1),
            (12, "Baela", 1, "F", "Teen", 1),
            (13, "Viserys", 1, "M", "Senior", 0),
            (14, "Alicent", 5, "F", "Adult", 1),
            (15, "Otto", 5, "M", "Senior", 1),
            (16, "Aegon", 1, "M", "Teen", 1),
            (17, "Cristen", 4, "M", "Adult", 1),
        ],
    )

    cursor.executemany(
        """
        INSERT INTO
            houses(id, name, reputation, is_noble)
        VALUES
            (?, ?, ?, ?);
        """,
        [
            (1, "Targaryen", 85, 1),
            (2, "Velaryon", 80, 1),
            (3, "Strong", 20, 0),
            (4, "Cole", -20, 0),
            (5, "Hightower", 50, 1),
        ],
    )

    cursor.executemany(
        """
        INSERT INTO
            relations (from_id, to_id, type)
        VALUES
            (?, ?, ?);
        """,
        [
            (4, 1, "Mother"),  # Jace -> Rhaenyra
            (4, 2, "Father"),  # Jace -> Laenor
            (4, 3, "BiologicalFather"),  # Jace -> Harwin
            (5, 6, "BiologicalFather"),  # Addam -> Corlys
            (2, 6, "BiologicalFather"),  # Laenor -> Corlys
            (2, 6, "Father"),  # Laenor -> Corlys
            (5, 7, "Mother"),  # Addam -> Marilda
            (8, 7, "Mother"),  # Alyn -> Marilda
            (8, 6, "BiologicalFather"),  # Alyn -> Corlys
            (2, 9, "Mother"),  # Laenor -> Rhaenys
            (10, 9, "Mother"),  # Laena -> Rhaenys
            (10, 6, "Father"),  # Laena -> Corlys
            (10, 6, "BiologicalFather"),  # Laena -> Corlys
            (6, 9, "Widower"),  # Corlys -> Rhaenys
            (6, 9, "FormerSpouse"),  # Corlys -> Rhaenys
            (9, 6, "FormerSpouse"),  # Rhaenys -> Corlys
            (12, 10, "Mother"),  # Baela -> Laena
            (12, 11, "Father"),  # Baela -> Daemon
            (12, 11, "BiologicalFather"),  # Baela -> Daemon
            (1, 11, "Spouse"),  # Rhaenyra -> Daemon
            (11, 1, "Spouse"),  # Daemon -> Rhaenyra
            (10, 11, "FormerSpouse"),  # Laena -> Daemon
            (11, 10, "FormerSpouse"),  # Daemon -> Laena
            (11, 10, "Widower"),  # Daemon -> Laena
            (1, 13, "Father"),  # Rhaenyra -> Viserys
            (1, 13, "BiologicalFather"),  # Rhaenyra -> Viserys
            (16, 14, "Mother"),  # Aegon -> Alicent
            (14, 15, "Father"),  # Alicent => Otto
            (14, 15, "BiologicalFather"),  # Alicent => Otto
        ],
    )

    db.commit()


def test_define_predicate_alias() -> None:
    """Ensure table aliases are registered and properly used."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    engine.execute_script(
        """
        ALIAS characters AS Character;
        """
    )

    rows = engine.query(
        """
        FIND ?x
        WHERE
            Character(id=?x)
        """,
        db,
    ).fetch_all()

    assert len(rows) == 17

    db.close()


def test_define_rule_alias() -> None:
    """Ensure rule aliases and registered and properly used."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    engine.execute_script(
        """
        ALIAS FromNobleHouse AS IsNobility;

        DEFINE
            FromNobleHouse(?x)
        WHERE
            characters(id=?x, house_id=?house_id)
            houses(id=?house_id, is_noble=TRUE);
        """
    )

    rows = engine.query(
        """
        FIND ?x
        WHERE
            IsNobility(x=?x);
        """,
        db,
    ).fetch_all()

    assert len(rows) == 12

    db.close()


def test_define_rule() -> None:
    """Ensure rules can be defined and used within a query."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    engine.execute_script(
        """
        DEFINE
            FromNobleHouse(?x)
        WHERE
            characters(id=?x, house_id=?house_id)
            houses(id=?house_id, is_noble=TRUE);
        """
    )

    rows = engine.query(
        """
        FIND ?x
        WHERE
            FromNobleHouse(x=?x);
        """,
        db,
    ).fetch_all()

    assert len(rows) == 12

    db.close()


def test_composite_rules() -> None:
    """Test composing a rule using other rules."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    engine.execute_script(
        """
        DEFINE
            IsAdult(?x)
        WHERE
            characters(id=?x, life_stage="Adult");

        DEFINE
            FromNobleHouse(?x)
        WHERE
            characters(id=?x, house_id=?house_id)
            houses(id=?house_id, is_noble=TRUE);

        DEFINE
            AdultNoble(?x)
        WHERE
            IsAdult(x=?x)
            FromNobleHouse(x=?x);
        """
    )

    rows = engine.query(
        """
        FIND
            ?x
        WHERE
            AdultNoble(x=?x);
        """,
        db,
    ).fetch_all()

    assert len(rows) == 7

    db.close()


def test_single_predicate_query() -> None:
    """Test queries composed of a single predicate."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND ?x
        WHERE
            characters(id=?x);
        """,
        db,
    ).fetch_all()

    assert len(rows) == 17

    rows = engine.query(
        """
        FIND ?x, ?house_id
        WHERE
            characters(id=?x, house_id=?house_id);
        """,
        db,
    ).fetch_all()

    assert len(rows) == 17

    db.close()


def test_query_output_aliases() -> None:
    """Test that queries output columns with the proper alias names."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            ?x AS character_id
        WHERE
            characters(id=?x);
        """,
        db,
    ).fetch_all()

    assert len(rows) == 17

    db.close()


def test_rule_param_aliases() -> None:
    """Test that rule parameters can be referred to using provided aliases."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    engine.execute_script(
        """
        DEFINE
            FromNobleHouse(?x AS character_id)
        WHERE
            characters(id=?x, house_id=?house_id)
            houses(id=?house_id, is_noble=TRUE);
        """
    )

    rows = engine.query(
        """
        FIND ?x
        WHERE
            FromNobleHouse(character_id=?x);
        """,
        db,
    ).fetch_all()

    assert len(rows) == 12

    db.close()


def test_multi_predicate_query() -> None:
    """Test a query that joins across multiple predicate statements."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            ?x
        WHERE
            characters(id=?x, house_id=?house_id)
            houses(id=?house_id, is_noble=TRUE);
        """,
        db,
    ).fetch_all()

    assert len(rows) == 12

    db.close()


def test_eq_filter() -> None:
    """Test the equals comparison filter."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            ?x
        WHERE
            characters(id=?x, life_stage=?life_stage, house_id=?house_id)
            houses(id=?house_id, name="Targaryen")
            (?life_stage = "Adult");
        """,
        db,
    ).fetch_all()

    assert len(rows) == 3

    db.close()


def test_neq_filter() -> None:
    """Test the not equals filter."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            ?x
        WHERE
            characters(id=?x, life_stage=?life_stage, house_id=?house_id)
            houses(id=?house_id, name="Targaryen")
            (?life_stage != "Adult");
        """,
        db,
    ).fetch_all()

    assert len(rows) == 3

    engine.execute_script(
        """
        ALIAS characters AS Character;
        ALIAS relations AS Relation;
        ALIAS houses AS House;

        DEFINE
            PaternalHalfSiblings(?x, ?y)
        WHERE
            Relation(from_id=?x, to_id=?bf, type="BiologicalFather")
            Relation(from_id=?y, to_id=?bf, type="BiologicalFather")
            Relation(from_id=?x, to_id=?x_m, type="Mother")
            Relation(from_id=?y, to_id=?y_m, type="Mother")
            ((?x_m != ?y_m) AND (?x != ?y));
        """
    )

    rows = engine.query(
        """
        FIND
        ?siblingId, ?siblingName
        WHERE
            Character(id=?adam_id, name="Addam")
            PaternalHalfSiblings(x=?adam_id, y=?siblingId)
            Character(id=?siblingId, name=?siblingName)
        ORDER BY ?siblingId;
        """,
        db,
    ).fetch_all()

    assert len(rows) == 2

    db.close()


def test_lt_filter() -> None:
    """Test the less than filter."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            ?house_id
        WHERE
            houses(id=?house_id, reputation=?rep)
            (?rep < 50);
        """,
        db,
    ).fetch_all()

    assert len(rows) == 2

    db.close()


def test_gt_filter() -> None:
    """Test the greater than filter."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            ?house_id
        WHERE
            houses(id=?house_id, reputation=?rep)
            (?rep > 50);
        """,
        db,
    ).fetch_all()

    assert len(rows) == 2

    db.close()


def test_lte_filter() -> None:
    """Test the less-than or equal to filter."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            ?house_id
        WHERE
            houses(id=?house_id, reputation=?rep)
            (?rep <= 50);
        """,
        db,
    ).fetch_all()

    assert len(rows) == 3

    db.close()


def test_gte_filter() -> None:
    """Test the greater-than or equal to filter."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            ?house_id
        WHERE
            houses(id=?house_id, reputation=?rep)
            (?rep >= 50);
        """,
        db,
    ).fetch_all()

    assert len(rows) == 3

    db.close()


def test_membership_filter() -> None:
    """Test the list membership filter."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            ?x
        WHERE
            characters(id=?x, house_id=?house_id)
            houses(id=?house_id, name=?family_name)
            (?family_name IN ["Velaryon", "Targaryen"]);
        """,
        db,
    ).fetch_all()

    assert len(rows) == 10

    rows = engine.query(
        """
        FIND
            ?x
        WHERE
            characters(id=?x, house_id=?house_id)
            houses(id=?house_id, name=?family_name)
            (?family_name NOT IN ["Velaryon", "Targaryen"]);
        """,
        db,
    ).fetch_all()

    assert len(rows) == 4

    db.close()


def test_null_check() -> None:
    """Test checking for NULL values."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            ?x
        WHERE
            characters(id=?x, house_id=NULL);
        """,
        db,
    ).fetch_all()

    assert len(rows) == 3

    rows = engine.query(
        """
        FIND
            ?x
        WHERE
            characters(id=?x, house_id=?house_id)
            (?house_id = NULL);
        """,
        db,
    ).fetch_all()

    assert len(rows) == 3

    rows = engine.query(
        """
        FIND
            ?x
        WHERE
            characters(id=?x, house_id=?house_id)
            (?house_id != NULL);
        """,
        db,
    ).fetch_all()

    assert len(rows) == 14

    db.close()


def test_and_statement() -> None:
    """Test using AND keyword to combine filter statements."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            ?x
        WHERE
            characters(id=?x, house_id=?house_id, is_alive=?is_alive)
            houses(id=?house_id, name=?family_name)
            ((?family_name = "Velaryon") AND (?is_alive = FALSE));
        """,
        db,
    ).fetch_all()

    assert len(rows) == 1

    db.close()


def test_or_statement() -> None:
    """Test using OR keyword to combine filter statements."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            ?x
        WHERE
            characters(id=?x, house_id=?house_id)
            houses(id=?house_id, name=?family_name)
            ((?family_name = "Velaryon") OR (?family_name = "Targaryen"));
        """,
        db,
    ).fetch_all()

    assert len(rows) == 10

    db.close()


def test_not_filter_statement() -> None:
    """Test using NOT keyword on filter statements."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            ?x
        WHERE
            characters(id=?x, house_id=?house_id)
            houses(id=?house_id, name=?family_name)
            (NOT (?family_name IN ["Velaryon", "Targaryen"]));
        """,
        db,
    ).fetch_all()

    assert len(rows) == 4

    rows = engine.query(
        """
        FIND
            ?x
        WHERE
            characters(id=?x, house_id=?house_id)
            houses(id=?house_id, name=?family_name)
            (NOT ((?family_name = "Velaryon") OR (?family_name = "Targaryen")));
        """,
        db,
    ).fetch_all()

    assert len(rows) == 4

    db.close()


def test_not_predicate_statement() -> None:
    """Test using NOT keyword on predicate statements."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            ?x
        WHERE
            characters(id=?x)
            NOT characters(id=?x, is_alive=FALSE);
        """,
        db,
    ).fetch_all()

    assert len(rows) == 13


def test_not_rule_statement() -> None:
    """Test using NOT keyword on rule statements."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    engine.execute_script(
        """
        DEFINE
            FromNobleHouse(?x)
        WHERE
            characters(id=?x, house_id=?house_id)
            houses(id=?house_id, is_noble=TRUE);
        """
    )

    rows = engine.query(
        """
        FIND ?x
        WHERE
            characters(id=?x)
            NOT FromNobleHouse(x=?x);
        """,
        db,
    ).fetch_all()

    assert len(rows) == 5

    db.close()


def test_alias_cycle_detection() -> None:
    """Test that alias cycles are detected."""

    aliases = {"A": "D", "B": "C", "G": "D", "E": "A", "F": "B"}

    assert has_alias_cycle(aliases) == (False, "")

    aliases["C"] = "F"

    assert has_alias_cycle(aliases) == (True, "B")


def test_duplicate_queries() -> None:
    """Test that running the same query twice does not throw duplicate table error."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    engine.execute_script(
        """
        ALIAS characters AS Character;
        ALIAS relations AS Relation;
        ALIAS houses AS House;

        DEFINE
            PaternalHalfSiblings(?x, ?y)
        WHERE
            Relation(from_id=?x, to_id=?bf, type="BiologicalFather")
            Relation(from_id=?y, to_id=?bf, type="BiologicalFather")
            Relation(from_id=?x, to_id=?x_m, type="Mother")
            Relation(from_id=?y, to_id=?y_m, type="Mother")
            ((?x_m != ?y_m) AND (?x != ?y));
        """
    )

    engine.query(
        """
        FIND
        ?siblingId, ?siblingName
        WHERE
            Character(id=?adam_id, name="Addam")
            PaternalHalfSiblings(x=?adam_id, y=?siblingId)
            Character(id=?siblingId, name=?siblingName)
        ORDER BY ?siblingId;
        """,
        db,
    )

    engine.query(
        """
        FIND
        ?siblingId, ?siblingName
        WHERE
            Character(id=?adam_id, name="Addam")
            PaternalHalfSiblings(x=?adam_id, y=?siblingId)
            Character(id=?siblingId, name=?siblingName)
        ORDER BY ?siblingId;
        """,
        db,
    )

    db.close()


def test_double_fetch_throws_error() -> None:
    """Ensure that double fetching data throws an exception."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    engine.execute_script(
        """
        ALIAS characters AS Character;
        ALIAS relations AS Relation;
        ALIAS houses AS House;

        DEFINE
            PaternalHalfSiblings(?x, ?y)
        WHERE
            Relation(from_id=?x, to_id=?bf, type="BiologicalFather")
            Relation(from_id=?y, to_id=?bf, type="BiologicalFather")
            Relation(from_id=?x, to_id=?x_m, type="Mother")
            Relation(from_id=?y, to_id=?y_m, type="Mother")
            ((?x_m != ?y_m) AND (?x != ?y));
        """
    )

    result = engine.query(
        """
        FIND
        ?siblingId, ?siblingName
        WHERE
            Character(id=?adam_id, name="Addam")
            PaternalHalfSiblings(x=?adam_id, y=?siblingId)
            Character(id=?siblingId, name=?siblingName)
        ORDER BY ?siblingId;
        """,
        db,
    )

    result.fetch_all()

    with pytest.raises(RuntimeError):
        result.fetch_all()

    result = engine.query(
        """
        FIND
        ?siblingId, ?siblingName
        WHERE
            Character(id=?adam_id, name="Addam")
            PaternalHalfSiblings(x=?adam_id, y=?siblingId)
            Character(id=?siblingId, name=?siblingName)
        ORDER BY ?siblingId;
        """,
        db,
    )

    # Have to call next() on the generator. Otherwise, the user
    # technically never read the data and it is still valid
    # to call fetch_all().
    next(result.fetch_chunks(20))

    with pytest.raises(RuntimeError):
        result.fetch_all()

    db.close()


def test_count_aggregate() -> None:
    """Test the COUNT aggregate function."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    engine.execute_script(
        """
        ALIAS characters AS Character;
        ALIAS relations AS Relation;
        ALIAS houses AS House;

        DEFINE
            HouseSize(?house_id AS id, COUNT(?character_id) AS size)
        WHERE
            Character(id=?character_id, house_id=?house_id)
            House(id=?house_id)
        GROUP BY ?house_id
        ORDER BY ?house_id ASC;
        """
    )

    rows = engine.query(
        """
        FIND
            ?house_id, ?size
        WHERE
            HouseSize(id=?house_id, size=?size)
        ORDER BY ?size DESC;
        """,
        db,
    ).fetch_all()

    assert len(rows) == 5
    assert rows[0] == (1, 6)
    assert rows[1] == (2, 4)
    assert rows[2] == (5, 2)

    db.close()


def test_order_by_asc() -> None:
    """Test ORDER BY ACS clause support."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            ?character_id, ?house_id
        WHERE
            characters(id=?character_id, house_id=?house_id)
        ORDER BY ?house_id;
        """,
        db,
    ).fetch_all()

    assert rows[0][1] is None
    assert rows[3][1] == 1
    assert rows[-1][1] == 5

    rows = engine.query(
        """
        FIND
            ?character_id, ?house_id
        WHERE
            characters(id=?character_id, house_id=?house_id)
        ORDER BY ?house_id ASC;
        """,
        db,
    ).fetch_all()

    assert rows[0][1] is None
    assert rows[3][1] == 1
    assert rows[-1][1] == 5

    rows = engine.query(
        """
        FIND
            ?character_id, ?house_id
        WHERE
            characters(id=?character_id, house_id=?house_id)
        ORDER BY ?house_id ASC NULLS FIRST;
        """,
        db,
    ).fetch_all()

    assert rows[0][1] is None
    assert rows[3][1] == 1
    assert rows[-1][1] == 5

    rows = engine.query(
        """
        FIND
            ?character_id, ?house_id
        WHERE
            characters(id=?character_id, house_id=?house_id)
        ORDER BY ?house_id ASC NULLS LAST;
        """,
        db,
    ).fetch_all()

    assert rows[0][1] == 1
    assert rows[6][1] == 2
    assert rows[-1][1] is None


def test_order_by_desc() -> None:
    """Test ORDER BY DESC clause support."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            ?character_id, ?house_id
        WHERE
            characters(id=?character_id, house_id=?house_id)
        ORDER BY ?house_id DESC;
        """,
        db,
    ).fetch_all()

    assert rows[0][1] == 5
    assert rows[2][1] == 4
    assert rows[-1][1] is None

    rows = engine.query(
        """
        FIND
            ?character_id, ?house_id
        WHERE
            characters(id=?character_id, house_id=?house_id)
        ORDER BY ?house_id DESC NULLS FIRST;
        """,
        db,
    ).fetch_all()

    assert rows[0][1] is None
    assert rows[3][1] == 5
    assert rows[-1][1] == 1

    rows = engine.query(
        """
        FIND
            ?character_id, ?house_id
        WHERE
            characters(id=?character_id, house_id=?house_id)
        ORDER BY ?house_id DESC NULLS LAST;
        """,
        db,
    ).fetch_all()

    assert rows[0][1] == 5
    assert rows[2][1] == 4
    assert rows[-1][1] is None


def test_order_by_multiple_columns() -> None:
    """Test ORDER BY support for multiple columns."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            ?character_id, ?name, ?house_id
        WHERE
            characters(id=?character_id, name=?name, house_id=?house_id)
        ORDER BY ?house_id ASC, ?name DESC;
        """,
        db,
    ).fetch_all()

    assert rows[0][1] == "Marilda"
    assert rows[1][1] == "Alyn"
    assert rows[2][1] == "Addam"
    assert rows[3][1] == "Viserys"


def test_limit() -> None:
    """Test LIMIT clause support."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            ?character_id, ?name
        WHERE
            characters(id=?character_id, name=?name)
        ORDER BY ?character_id
        LIMIT 5;
        """,
        db,
    ).fetch_all()

    assert len(rows) == 5
    assert rows[0][1] == "Rhaenyra"
    assert rows[1][1] == "Laenor"
    assert rows[2][1] == "Harwin"
    assert rows[3][1] == "Jacaerys"
    assert rows[4][1] == "Addam"


def test_limit_offset() -> None:
    """Test LIMIT with OFFSET support."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            ?character_id, ?name
        WHERE
            characters(id=?character_id, name=?name)
        ORDER BY ?character_id
        LIMIT 5 OFFSET 5;
        """,
        db,
    ).fetch_all()

    assert len(rows) == 5
    assert rows[0][1] == "Corlys"
    assert rows[1][1] == "Marilda"
    assert rows[2][1] == "Alyn"
    assert rows[3][1] == "Rhaenys"
    assert rows[4][1] == "Laena"


def test_group_by() -> None:
    """Test GROUP BY support."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            ?house_id, COUNT(?character_id) AS size
        WHERE
            characters(id=?character_id, house_id=?house_id)
        GROUP BY ?house_id
        ORDER BY ?size DESC;
        """,
        db,
    ).fetch_all()

    assert len(rows) == 6


def test_group_by_multiple_columns() -> None:
    """Test GROUP BY multiple columns support."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            ?house_id, ?sex, COUNT(?character_id) AS size
        WHERE
            characters(id=?character_id, sex=?sex, house_id=?house_id)
        GROUP BY ?house_id, ?sex
        ORDER BY ?size DESC;
        """,
        db,
    ).fetch_all()

    assert len(rows) == 10


def test_avg_aggregate() -> None:
    """Test support for the AVG aggregate function."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            AVG(?rep) AS avg_rep
        WHERE
            houses(reputation=?rep);
        """,
        db,
    ).fetch_all()

    assert rows[0][0] == 43


def test_max_aggregate() -> None:
    """Test support for the MAX aggregate function."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            MAX(?rep)
        WHERE
            houses(reputation=?rep);
        """,
        db,
    ).fetch_all()

    assert rows[0][0] == 85


def test_min_aggregate() -> None:
    """Test support for the MIN aggregate function."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            MIN(?rep)
        WHERE
            houses(reputation=?rep);
        """,
        db,
    ).fetch_all()

    assert rows[0][0] == -20


def test_sum_aggregate() -> None:
    """Test support for the SUM aggregate function."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            SUM(?rep)
        WHERE
            houses(reputation=?rep);
        """,
        db,
    ).fetch_all()

    assert rows[0][0] == 215


def test_query_bindings() -> None:
    """Test passing bindings to queries."""

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            ?x, ?y
        WHERE
            characters(id=?x)
            characters(id=?y)
            (?x != ?y);
        """,
        db,
        {"?x": 1},
    ).fetch_all()

    assert len(rows) == 16

    db = sqlite3.Connection(":memory:")

    initialize_test_data(db)

    engine = drolta.engine.QueryEngine()

    rows = engine.query(
        """
        FIND
            ?x, ?y
        WHERE
            characters(id=?x, name=?name)
            characters(id=?y)
            (?x != ?y);
        """,
        db,
        {"?name": "Rhaenyra", "?y": 2},
    ).fetch_all()

    assert len(rows) == 1

    db.close()



# Generated from ./Drolta.g4 by ANTLR 4.13.2
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,45,261,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,2,4,7,4,2,5,7,5,2,6,7,
        6,2,7,7,7,2,8,7,8,2,9,7,9,2,10,7,10,2,11,7,11,2,12,7,12,2,13,7,13,
        2,14,7,14,2,15,7,15,2,16,7,16,2,17,7,17,2,18,7,18,2,19,7,19,2,20,
        7,20,1,0,1,0,1,0,5,0,46,8,0,10,0,12,0,49,9,0,3,0,51,8,0,1,0,3,0,
        54,8,0,1,0,1,0,1,1,1,1,1,1,3,1,61,8,1,1,2,1,2,1,2,1,2,1,2,1,3,1,
        3,1,3,3,3,71,8,3,1,3,3,3,74,8,3,1,3,3,3,77,8,3,1,4,1,4,1,4,1,4,1,
        4,1,4,5,4,85,8,4,10,4,12,4,88,9,4,3,4,90,8,4,1,4,1,4,1,5,1,5,1,5,
        3,5,97,8,5,1,5,3,5,100,8,5,1,5,3,5,103,8,5,1,6,1,6,1,6,1,6,5,6,109,
        8,6,10,6,12,6,112,9,6,3,6,114,8,6,1,7,1,7,1,7,1,7,1,7,1,7,3,7,122,
        8,7,1,7,1,7,1,7,3,7,127,8,7,3,7,129,8,7,1,8,1,8,1,8,5,8,134,8,8,
        10,8,12,8,137,9,8,3,8,139,8,8,1,9,1,9,1,9,1,9,1,9,5,9,146,8,9,10,
        9,12,9,149,9,9,1,10,1,10,3,10,153,8,10,1,10,1,10,3,10,157,8,10,1,
        11,1,11,1,11,1,11,1,11,5,11,164,8,11,10,11,12,11,167,9,11,1,12,1,
        12,1,12,1,12,3,12,173,8,12,1,13,1,13,3,13,177,8,13,1,14,1,14,1,14,
        1,14,1,14,1,14,1,14,1,14,1,14,3,14,188,8,14,1,14,1,14,1,14,1,14,
        1,14,1,14,1,14,1,14,1,14,1,14,1,14,1,14,1,14,1,14,1,14,1,14,1,14,
        1,14,1,14,1,14,1,14,3,14,211,8,14,1,15,1,15,1,15,1,15,5,15,217,8,
        15,10,15,12,15,220,9,15,3,15,222,8,15,1,15,1,15,1,16,1,16,1,17,1,
        17,1,17,1,17,1,17,5,17,233,8,17,10,17,12,17,236,9,17,3,17,238,8,
        17,1,17,1,17,1,17,3,17,243,8,17,1,18,1,18,1,18,1,18,1,19,1,19,1,
        19,1,19,1,19,1,19,1,19,3,19,256,8,19,1,20,1,20,1,20,1,20,0,0,21,
        0,2,4,6,8,10,12,14,16,18,20,22,24,26,28,30,32,34,36,38,40,0,3,1,
        0,30,31,1,0,33,34,1,0,24,29,281,0,50,1,0,0,0,2,60,1,0,0,0,4,62,1,
        0,0,0,6,67,1,0,0,0,8,78,1,0,0,0,10,93,1,0,0,0,12,104,1,0,0,0,14,
        128,1,0,0,0,16,130,1,0,0,0,18,140,1,0,0,0,20,150,1,0,0,0,22,158,
        1,0,0,0,24,168,1,0,0,0,26,176,1,0,0,0,28,210,1,0,0,0,30,212,1,0,
        0,0,32,225,1,0,0,0,34,242,1,0,0,0,36,244,1,0,0,0,38,255,1,0,0,0,
        40,257,1,0,0,0,42,47,3,2,1,0,43,44,5,2,0,0,44,46,3,2,1,0,45,43,1,
        0,0,0,46,49,1,0,0,0,47,45,1,0,0,0,47,48,1,0,0,0,48,51,1,0,0,0,49,
        47,1,0,0,0,50,42,1,0,0,0,50,51,1,0,0,0,51,53,1,0,0,0,52,54,5,2,0,
        0,53,52,1,0,0,0,53,54,1,0,0,0,54,55,1,0,0,0,55,56,5,0,0,1,56,1,1,
        0,0,0,57,61,3,4,2,0,58,61,3,6,3,0,59,61,3,10,5,0,60,57,1,0,0,0,60,
        58,1,0,0,0,60,59,1,0,0,0,61,3,1,0,0,0,62,63,5,22,0,0,63,64,5,39,
        0,0,64,65,5,23,0,0,65,66,5,39,0,0,66,5,1,0,0,0,67,68,3,8,4,0,68,
        70,3,16,8,0,69,71,3,22,11,0,70,69,1,0,0,0,70,71,1,0,0,0,71,73,1,
        0,0,0,72,74,3,18,9,0,73,72,1,0,0,0,73,74,1,0,0,0,74,76,1,0,0,0,75,
        77,3,24,12,0,76,75,1,0,0,0,76,77,1,0,0,0,77,7,1,0,0,0,78,79,5,21,
        0,0,79,80,5,39,0,0,80,89,5,4,0,0,81,86,3,14,7,0,82,83,5,9,0,0,83,
        85,3,14,7,0,84,82,1,0,0,0,85,88,1,0,0,0,86,84,1,0,0,0,86,87,1,0,
        0,0,87,90,1,0,0,0,88,86,1,0,0,0,89,81,1,0,0,0,89,90,1,0,0,0,90,91,
        1,0,0,0,91,92,5,5,0,0,92,9,1,0,0,0,93,94,3,12,6,0,94,96,3,16,8,0,
        95,97,3,22,11,0,96,95,1,0,0,0,96,97,1,0,0,0,97,99,1,0,0,0,98,100,
        3,18,9,0,99,98,1,0,0,0,99,100,1,0,0,0,100,102,1,0,0,0,101,103,3,
        24,12,0,102,101,1,0,0,0,102,103,1,0,0,0,103,11,1,0,0,0,104,113,5,
        10,0,0,105,110,3,14,7,0,106,107,5,9,0,0,107,109,3,14,7,0,108,106,
        1,0,0,0,109,112,1,0,0,0,110,108,1,0,0,0,110,111,1,0,0,0,111,114,
        1,0,0,0,112,110,1,0,0,0,113,105,1,0,0,0,113,114,1,0,0,0,114,13,1,
        0,0,0,115,116,5,39,0,0,116,117,5,4,0,0,117,118,3,40,20,0,118,121,
        5,5,0,0,119,120,5,23,0,0,120,122,5,39,0,0,121,119,1,0,0,0,121,122,
        1,0,0,0,122,129,1,0,0,0,123,126,3,40,20,0,124,125,5,23,0,0,125,127,
        5,39,0,0,126,124,1,0,0,0,126,127,1,0,0,0,127,129,1,0,0,0,128,115,
        1,0,0,0,128,123,1,0,0,0,129,15,1,0,0,0,130,138,5,11,0,0,131,135,
        3,26,13,0,132,134,3,26,13,0,133,132,1,0,0,0,134,137,1,0,0,0,135,
        133,1,0,0,0,135,136,1,0,0,0,136,139,1,0,0,0,137,135,1,0,0,0,138,
        131,1,0,0,0,138,139,1,0,0,0,139,17,1,0,0,0,140,141,5,18,0,0,141,
        142,5,16,0,0,142,147,3,20,10,0,143,144,5,9,0,0,144,146,3,20,10,0,
        145,143,1,0,0,0,146,149,1,0,0,0,147,145,1,0,0,0,147,148,1,0,0,0,
        148,19,1,0,0,0,149,147,1,0,0,0,150,152,3,40,20,0,151,153,7,0,0,0,
        152,151,1,0,0,0,152,153,1,0,0,0,153,156,1,0,0,0,154,155,5,35,0,0,
        155,157,7,1,0,0,156,154,1,0,0,0,156,157,1,0,0,0,157,21,1,0,0,0,158,
        159,5,19,0,0,159,160,5,16,0,0,160,165,3,40,20,0,161,162,5,9,0,0,
        162,164,3,40,20,0,163,161,1,0,0,0,164,167,1,0,0,0,165,163,1,0,0,
        0,165,166,1,0,0,0,166,23,1,0,0,0,167,165,1,0,0,0,168,169,5,15,0,
        0,169,172,5,40,0,0,170,171,5,32,0,0,171,173,5,40,0,0,172,170,1,0,
        0,0,172,173,1,0,0,0,173,25,1,0,0,0,174,177,3,28,14,0,175,177,3,34,
        17,0,176,174,1,0,0,0,176,175,1,0,0,0,177,27,1,0,0,0,178,179,5,4,
        0,0,179,180,3,40,20,0,180,181,3,32,16,0,181,182,3,38,19,0,182,183,
        5,5,0,0,183,211,1,0,0,0,184,185,5,4,0,0,185,187,3,40,20,0,186,188,
        5,14,0,0,187,186,1,0,0,0,187,188,1,0,0,0,188,189,1,0,0,0,189,190,
        5,17,0,0,190,191,3,30,15,0,191,192,5,5,0,0,192,211,1,0,0,0,193,194,
        5,4,0,0,194,195,3,28,14,0,195,196,5,13,0,0,196,197,3,28,14,0,197,
        198,5,5,0,0,198,211,1,0,0,0,199,200,5,4,0,0,200,201,3,28,14,0,201,
        202,5,12,0,0,202,203,3,28,14,0,203,204,5,5,0,0,204,211,1,0,0,0,205,
        206,5,4,0,0,206,207,5,14,0,0,207,208,3,28,14,0,208,209,5,5,0,0,209,
        211,1,0,0,0,210,178,1,0,0,0,210,184,1,0,0,0,210,193,1,0,0,0,210,
        199,1,0,0,0,210,205,1,0,0,0,211,29,1,0,0,0,212,221,5,6,0,0,213,218,
        3,38,19,0,214,215,5,9,0,0,215,217,3,38,19,0,216,214,1,0,0,0,217,
        220,1,0,0,0,218,216,1,0,0,0,218,219,1,0,0,0,219,222,1,0,0,0,220,
        218,1,0,0,0,221,213,1,0,0,0,221,222,1,0,0,0,222,223,1,0,0,0,223,
        224,5,7,0,0,224,31,1,0,0,0,225,226,7,2,0,0,226,33,1,0,0,0,227,228,
        5,39,0,0,228,237,5,4,0,0,229,234,3,36,18,0,230,231,5,9,0,0,231,233,
        3,36,18,0,232,230,1,0,0,0,233,236,1,0,0,0,234,232,1,0,0,0,234,235,
        1,0,0,0,235,238,1,0,0,0,236,234,1,0,0,0,237,229,1,0,0,0,237,238,
        1,0,0,0,238,239,1,0,0,0,239,243,5,5,0,0,240,241,5,14,0,0,241,243,
        3,34,17,0,242,227,1,0,0,0,242,240,1,0,0,0,243,35,1,0,0,0,244,245,
        5,39,0,0,245,246,5,24,0,0,246,247,3,38,19,0,247,37,1,0,0,0,248,256,
        3,40,20,0,249,256,5,40,0,0,250,256,5,41,0,0,251,256,5,42,0,0,252,
        256,5,36,0,0,253,256,5,37,0,0,254,256,5,38,0,0,255,248,1,0,0,0,255,
        249,1,0,0,0,255,250,1,0,0,0,255,251,1,0,0,0,255,252,1,0,0,0,255,
        253,1,0,0,0,255,254,1,0,0,0,256,39,1,0,0,0,257,258,5,1,0,0,258,259,
        5,39,0,0,259,41,1,0,0,0,33,47,50,53,60,70,73,76,86,89,96,99,102,
        110,113,121,126,128,135,138,147,152,156,165,172,176,187,210,218,
        221,234,237,242,255
    ]

class DroltaParser ( Parser ):

    grammarFileName = "Drolta.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'?'", "';'", "'.'", "'('", "')'", "'['", 
                     "']'", "'::'", "','", "'FIND'", "'WHERE'", "'OR'", 
                     "'AND'", "'NOT'", "'LIMIT'", "'BY'", "'IN'", "'ORDER'", 
                     "'GROUP'", "'USING'", "'DEFINE'", "'ALIAS'", "'AS'", 
                     "'='", "'!='", "'<='", "'<'", "'>='", "'>'", "'ASC'", 
                     "'DESC'", "'OFFSET'", "'FIRST'", "'LAST'", "'NULLS'", 
                     "'NULL'", "'TRUE'", "'FALSE'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "S_COL", "DOT", "OPEN_PAR", 
                      "CLOSE_PAR", "BRACKET_L", "BRACKET_R", "DBL_COL", 
                      "COMMA", "FIND", "WHERE", "OR", "AND", "NOT", "LIMIT", 
                      "BY", "IN", "ORDER", "GROUP", "USING", "DEFINE", "ALIAS", 
                      "AS", "EQ", "NEQ", "LTE", "LT", "GTE", "GT", "ASC", 
                      "DESC", "OFFSET", "FIRST", "LAST", "NULLS", "NULL", 
                      "TRUE", "FALSE", "IDENTIFIER", "INT_LITERAL", "FLOAT_LITERAL", 
                      "STRING_LITERAL", "SINGLE_LINE_COMMENT", "MULTILINE_COMMENT", 
                      "WHITESPACE" ]

    RULE_prog = 0
    RULE_prog_statement = 1
    RULE_alias_declaration = 2
    RULE_rule_declaration = 3
    RULE_define_clause = 4
    RULE_query = 5
    RULE_find_clause = 6
    RULE_result_var = 7
    RULE_where_clause = 8
    RULE_order_by_statement = 9
    RULE_ordering_term = 10
    RULE_group_by_statement = 11
    RULE_limit_statement = 12
    RULE_where_statement = 13
    RULE_filter_statement = 14
    RULE_atom_list = 15
    RULE_comparison_operator = 16
    RULE_predicate_statement = 17
    RULE_predicate_param = 18
    RULE_atom = 19
    RULE_variable = 20

    ruleNames =  [ "prog", "prog_statement", "alias_declaration", "rule_declaration", 
                   "define_clause", "query", "find_clause", "result_var", 
                   "where_clause", "order_by_statement", "ordering_term", 
                   "group_by_statement", "limit_statement", "where_statement", 
                   "filter_statement", "atom_list", "comparison_operator", 
                   "predicate_statement", "predicate_param", "atom", "variable" ]

    EOF = Token.EOF
    T__0=1
    S_COL=2
    DOT=3
    OPEN_PAR=4
    CLOSE_PAR=5
    BRACKET_L=6
    BRACKET_R=7
    DBL_COL=8
    COMMA=9
    FIND=10
    WHERE=11
    OR=12
    AND=13
    NOT=14
    LIMIT=15
    BY=16
    IN=17
    ORDER=18
    GROUP=19
    USING=20
    DEFINE=21
    ALIAS=22
    AS=23
    EQ=24
    NEQ=25
    LTE=26
    LT=27
    GTE=28
    GT=29
    ASC=30
    DESC=31
    OFFSET=32
    FIRST=33
    LAST=34
    NULLS=35
    NULL=36
    TRUE=37
    FALSE=38
    IDENTIFIER=39
    INT_LITERAL=40
    FLOAT_LITERAL=41
    STRING_LITERAL=42
    SINGLE_LINE_COMMENT=43
    MULTILINE_COMMENT=44
    WHITESPACE=45

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.2")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class ProgContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EOF(self):
            return self.getToken(DroltaParser.EOF, 0)

        def prog_statement(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DroltaParser.Prog_statementContext)
            else:
                return self.getTypedRuleContext(DroltaParser.Prog_statementContext,i)


        def S_COL(self, i:int=None):
            if i is None:
                return self.getTokens(DroltaParser.S_COL)
            else:
                return self.getToken(DroltaParser.S_COL, i)

        def getRuleIndex(self):
            return DroltaParser.RULE_prog

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterProg" ):
                listener.enterProg(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitProg" ):
                listener.exitProg(self)




    def prog(self):

        localctx = DroltaParser.ProgContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_prog)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 50
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la) & ~0x3f) == 0 and ((1 << _la) & 6292480) != 0):
                self.state = 42
                self.prog_statement()
                self.state = 47
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,0,self._ctx)
                while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                    if _alt==1:
                        self.state = 43
                        self.match(DroltaParser.S_COL)
                        self.state = 44
                        self.prog_statement() 
                    self.state = 49
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input,0,self._ctx)



            self.state = 53
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==2:
                self.state = 52
                self.match(DroltaParser.S_COL)


            self.state = 55
            self.match(DroltaParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Prog_statementContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def alias_declaration(self):
            return self.getTypedRuleContext(DroltaParser.Alias_declarationContext,0)


        def rule_declaration(self):
            return self.getTypedRuleContext(DroltaParser.Rule_declarationContext,0)


        def query(self):
            return self.getTypedRuleContext(DroltaParser.QueryContext,0)


        def getRuleIndex(self):
            return DroltaParser.RULE_prog_statement

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterProg_statement" ):
                listener.enterProg_statement(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitProg_statement" ):
                listener.exitProg_statement(self)




    def prog_statement(self):

        localctx = DroltaParser.Prog_statementContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_prog_statement)
        try:
            self.state = 60
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [22]:
                self.enterOuterAlt(localctx, 1)
                self.state = 57
                self.alias_declaration()
                pass
            elif token in [21]:
                self.enterOuterAlt(localctx, 2)
                self.state = 58
                self.rule_declaration()
                pass
            elif token in [10]:
                self.enterOuterAlt(localctx, 3)
                self.state = 59
                self.query()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Alias_declarationContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.original = None # Token
            self.alias = None # Token

        def ALIAS(self):
            return self.getToken(DroltaParser.ALIAS, 0)

        def AS(self):
            return self.getToken(DroltaParser.AS, 0)

        def IDENTIFIER(self, i:int=None):
            if i is None:
                return self.getTokens(DroltaParser.IDENTIFIER)
            else:
                return self.getToken(DroltaParser.IDENTIFIER, i)

        def getRuleIndex(self):
            return DroltaParser.RULE_alias_declaration

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAlias_declaration" ):
                listener.enterAlias_declaration(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAlias_declaration" ):
                listener.exitAlias_declaration(self)




    def alias_declaration(self):

        localctx = DroltaParser.Alias_declarationContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_alias_declaration)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 62
            self.match(DroltaParser.ALIAS)
            self.state = 63
            localctx.original = self.match(DroltaParser.IDENTIFIER)
            self.state = 64
            self.match(DroltaParser.AS)
            self.state = 65
            localctx.alias = self.match(DroltaParser.IDENTIFIER)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Rule_declarationContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def define_clause(self):
            return self.getTypedRuleContext(DroltaParser.Define_clauseContext,0)


        def where_clause(self):
            return self.getTypedRuleContext(DroltaParser.Where_clauseContext,0)


        def group_by_statement(self):
            return self.getTypedRuleContext(DroltaParser.Group_by_statementContext,0)


        def order_by_statement(self):
            return self.getTypedRuleContext(DroltaParser.Order_by_statementContext,0)


        def limit_statement(self):
            return self.getTypedRuleContext(DroltaParser.Limit_statementContext,0)


        def getRuleIndex(self):
            return DroltaParser.RULE_rule_declaration

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterRule_declaration" ):
                listener.enterRule_declaration(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitRule_declaration" ):
                listener.exitRule_declaration(self)




    def rule_declaration(self):

        localctx = DroltaParser.Rule_declarationContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_rule_declaration)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 67
            self.define_clause()
            self.state = 68
            self.where_clause()
            self.state = 70
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==19:
                self.state = 69
                self.group_by_statement()


            self.state = 73
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==18:
                self.state = 72
                self.order_by_statement()


            self.state = 76
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==15:
                self.state = 75
                self.limit_statement()


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Define_clauseContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.ruleName = None # Token

        def DEFINE(self):
            return self.getToken(DroltaParser.DEFINE, 0)

        def OPEN_PAR(self):
            return self.getToken(DroltaParser.OPEN_PAR, 0)

        def CLOSE_PAR(self):
            return self.getToken(DroltaParser.CLOSE_PAR, 0)

        def IDENTIFIER(self):
            return self.getToken(DroltaParser.IDENTIFIER, 0)

        def result_var(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DroltaParser.Result_varContext)
            else:
                return self.getTypedRuleContext(DroltaParser.Result_varContext,i)


        def COMMA(self, i:int=None):
            if i is None:
                return self.getTokens(DroltaParser.COMMA)
            else:
                return self.getToken(DroltaParser.COMMA, i)

        def getRuleIndex(self):
            return DroltaParser.RULE_define_clause

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterDefine_clause" ):
                listener.enterDefine_clause(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitDefine_clause" ):
                listener.exitDefine_clause(self)




    def define_clause(self):

        localctx = DroltaParser.Define_clauseContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_define_clause)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 78
            self.match(DroltaParser.DEFINE)
            self.state = 79
            localctx.ruleName = self.match(DroltaParser.IDENTIFIER)
            self.state = 80
            self.match(DroltaParser.OPEN_PAR)
            self.state = 89
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==1 or _la==39:
                self.state = 81
                self.result_var()
                self.state = 86
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==9:
                    self.state = 82
                    self.match(DroltaParser.COMMA)
                    self.state = 83
                    self.result_var()
                    self.state = 88
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)



            self.state = 91
            self.match(DroltaParser.CLOSE_PAR)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class QueryContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def find_clause(self):
            return self.getTypedRuleContext(DroltaParser.Find_clauseContext,0)


        def where_clause(self):
            return self.getTypedRuleContext(DroltaParser.Where_clauseContext,0)


        def group_by_statement(self):
            return self.getTypedRuleContext(DroltaParser.Group_by_statementContext,0)


        def order_by_statement(self):
            return self.getTypedRuleContext(DroltaParser.Order_by_statementContext,0)


        def limit_statement(self):
            return self.getTypedRuleContext(DroltaParser.Limit_statementContext,0)


        def getRuleIndex(self):
            return DroltaParser.RULE_query

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterQuery" ):
                listener.enterQuery(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitQuery" ):
                listener.exitQuery(self)




    def query(self):

        localctx = DroltaParser.QueryContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_query)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 93
            self.find_clause()
            self.state = 94
            self.where_clause()
            self.state = 96
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==19:
                self.state = 95
                self.group_by_statement()


            self.state = 99
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==18:
                self.state = 98
                self.order_by_statement()


            self.state = 102
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==15:
                self.state = 101
                self.limit_statement()


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Find_clauseContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def FIND(self):
            return self.getToken(DroltaParser.FIND, 0)

        def result_var(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DroltaParser.Result_varContext)
            else:
                return self.getTypedRuleContext(DroltaParser.Result_varContext,i)


        def COMMA(self, i:int=None):
            if i is None:
                return self.getTokens(DroltaParser.COMMA)
            else:
                return self.getToken(DroltaParser.COMMA, i)

        def getRuleIndex(self):
            return DroltaParser.RULE_find_clause

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterFind_clause" ):
                listener.enterFind_clause(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitFind_clause" ):
                listener.exitFind_clause(self)




    def find_clause(self):

        localctx = DroltaParser.Find_clauseContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_find_clause)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 104
            self.match(DroltaParser.FIND)
            self.state = 113
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==1 or _la==39:
                self.state = 105
                self.result_var()
                self.state = 110
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==9:
                    self.state = 106
                    self.match(DroltaParser.COMMA)
                    self.state = 107
                    self.result_var()
                    self.state = 112
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)



        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Result_varContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.aggregateName = None # Token
            self.alias = None # Token

        def OPEN_PAR(self):
            return self.getToken(DroltaParser.OPEN_PAR, 0)

        def variable(self):
            return self.getTypedRuleContext(DroltaParser.VariableContext,0)


        def CLOSE_PAR(self):
            return self.getToken(DroltaParser.CLOSE_PAR, 0)

        def IDENTIFIER(self, i:int=None):
            if i is None:
                return self.getTokens(DroltaParser.IDENTIFIER)
            else:
                return self.getToken(DroltaParser.IDENTIFIER, i)

        def AS(self):
            return self.getToken(DroltaParser.AS, 0)

        def getRuleIndex(self):
            return DroltaParser.RULE_result_var

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterResult_var" ):
                listener.enterResult_var(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitResult_var" ):
                listener.exitResult_var(self)




    def result_var(self):

        localctx = DroltaParser.Result_varContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_result_var)
        self._la = 0 # Token type
        try:
            self.state = 128
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [39]:
                self.enterOuterAlt(localctx, 1)
                self.state = 115
                localctx.aggregateName = self.match(DroltaParser.IDENTIFIER)
                self.state = 116
                self.match(DroltaParser.OPEN_PAR)
                self.state = 117
                self.variable()
                self.state = 118
                self.match(DroltaParser.CLOSE_PAR)
                self.state = 121
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==23:
                    self.state = 119
                    self.match(DroltaParser.AS)
                    self.state = 120
                    localctx.alias = self.match(DroltaParser.IDENTIFIER)


                pass
            elif token in [1]:
                self.enterOuterAlt(localctx, 2)
                self.state = 123
                self.variable()
                self.state = 126
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==23:
                    self.state = 124
                    self.match(DroltaParser.AS)
                    self.state = 125
                    localctx.alias = self.match(DroltaParser.IDENTIFIER)


                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Where_clauseContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def WHERE(self):
            return self.getToken(DroltaParser.WHERE, 0)

        def where_statement(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DroltaParser.Where_statementContext)
            else:
                return self.getTypedRuleContext(DroltaParser.Where_statementContext,i)


        def getRuleIndex(self):
            return DroltaParser.RULE_where_clause

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterWhere_clause" ):
                listener.enterWhere_clause(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitWhere_clause" ):
                listener.exitWhere_clause(self)




    def where_clause(self):

        localctx = DroltaParser.Where_clauseContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_where_clause)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 130
            self.match(DroltaParser.WHERE)
            self.state = 138
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la) & ~0x3f) == 0 and ((1 << _la) & 549755830288) != 0):
                self.state = 131
                self.where_statement()
                self.state = 135
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while (((_la) & ~0x3f) == 0 and ((1 << _la) & 549755830288) != 0):
                    self.state = 132
                    self.where_statement()
                    self.state = 137
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)



        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Order_by_statementContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ORDER(self):
            return self.getToken(DroltaParser.ORDER, 0)

        def BY(self):
            return self.getToken(DroltaParser.BY, 0)

        def ordering_term(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DroltaParser.Ordering_termContext)
            else:
                return self.getTypedRuleContext(DroltaParser.Ordering_termContext,i)


        def COMMA(self, i:int=None):
            if i is None:
                return self.getTokens(DroltaParser.COMMA)
            else:
                return self.getToken(DroltaParser.COMMA, i)

        def getRuleIndex(self):
            return DroltaParser.RULE_order_by_statement

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterOrder_by_statement" ):
                listener.enterOrder_by_statement(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitOrder_by_statement" ):
                listener.exitOrder_by_statement(self)




    def order_by_statement(self):

        localctx = DroltaParser.Order_by_statementContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_order_by_statement)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 140
            self.match(DroltaParser.ORDER)
            self.state = 141
            self.match(DroltaParser.BY)
            self.state = 142
            self.ordering_term()
            self.state = 147
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==9:
                self.state = 143
                self.match(DroltaParser.COMMA)
                self.state = 144
                self.ordering_term()
                self.state = 149
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Ordering_termContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def variable(self):
            return self.getTypedRuleContext(DroltaParser.VariableContext,0)


        def NULLS(self):
            return self.getToken(DroltaParser.NULLS, 0)

        def ASC(self):
            return self.getToken(DroltaParser.ASC, 0)

        def DESC(self):
            return self.getToken(DroltaParser.DESC, 0)

        def FIRST(self):
            return self.getToken(DroltaParser.FIRST, 0)

        def LAST(self):
            return self.getToken(DroltaParser.LAST, 0)

        def getRuleIndex(self):
            return DroltaParser.RULE_ordering_term

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterOrdering_term" ):
                listener.enterOrdering_term(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitOrdering_term" ):
                listener.exitOrdering_term(self)




    def ordering_term(self):

        localctx = DroltaParser.Ordering_termContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_ordering_term)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 150
            self.variable()
            self.state = 152
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==30 or _la==31:
                self.state = 151
                _la = self._input.LA(1)
                if not(_la==30 or _la==31):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()


            self.state = 156
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==35:
                self.state = 154
                self.match(DroltaParser.NULLS)
                self.state = 155
                _la = self._input.LA(1)
                if not(_la==33 or _la==34):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Group_by_statementContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def GROUP(self):
            return self.getToken(DroltaParser.GROUP, 0)

        def BY(self):
            return self.getToken(DroltaParser.BY, 0)

        def variable(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DroltaParser.VariableContext)
            else:
                return self.getTypedRuleContext(DroltaParser.VariableContext,i)


        def COMMA(self, i:int=None):
            if i is None:
                return self.getTokens(DroltaParser.COMMA)
            else:
                return self.getToken(DroltaParser.COMMA, i)

        def getRuleIndex(self):
            return DroltaParser.RULE_group_by_statement

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterGroup_by_statement" ):
                listener.enterGroup_by_statement(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitGroup_by_statement" ):
                listener.exitGroup_by_statement(self)




    def group_by_statement(self):

        localctx = DroltaParser.Group_by_statementContext(self, self._ctx, self.state)
        self.enterRule(localctx, 22, self.RULE_group_by_statement)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 158
            self.match(DroltaParser.GROUP)
            self.state = 159
            self.match(DroltaParser.BY)
            self.state = 160
            self.variable()
            self.state = 165
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==9:
                self.state = 161
                self.match(DroltaParser.COMMA)
                self.state = 162
                self.variable()
                self.state = 167
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Limit_statementContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.limitVal = None # Token
            self.offsetVal = None # Token

        def LIMIT(self):
            return self.getToken(DroltaParser.LIMIT, 0)

        def INT_LITERAL(self, i:int=None):
            if i is None:
                return self.getTokens(DroltaParser.INT_LITERAL)
            else:
                return self.getToken(DroltaParser.INT_LITERAL, i)

        def OFFSET(self):
            return self.getToken(DroltaParser.OFFSET, 0)

        def getRuleIndex(self):
            return DroltaParser.RULE_limit_statement

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterLimit_statement" ):
                listener.enterLimit_statement(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitLimit_statement" ):
                listener.exitLimit_statement(self)




    def limit_statement(self):

        localctx = DroltaParser.Limit_statementContext(self, self._ctx, self.state)
        self.enterRule(localctx, 24, self.RULE_limit_statement)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 168
            self.match(DroltaParser.LIMIT)
            self.state = 169
            localctx.limitVal = self.match(DroltaParser.INT_LITERAL)
            self.state = 172
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==32:
                self.state = 170
                self.match(DroltaParser.OFFSET)
                self.state = 171
                localctx.offsetVal = self.match(DroltaParser.INT_LITERAL)


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Where_statementContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def filter_statement(self):
            return self.getTypedRuleContext(DroltaParser.Filter_statementContext,0)


        def predicate_statement(self):
            return self.getTypedRuleContext(DroltaParser.Predicate_statementContext,0)


        def getRuleIndex(self):
            return DroltaParser.RULE_where_statement

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterWhere_statement" ):
                listener.enterWhere_statement(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitWhere_statement" ):
                listener.exitWhere_statement(self)




    def where_statement(self):

        localctx = DroltaParser.Where_statementContext(self, self._ctx, self.state)
        self.enterRule(localctx, 26, self.RULE_where_statement)
        try:
            self.state = 176
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [4]:
                self.enterOuterAlt(localctx, 1)
                self.state = 174
                self.filter_statement()
                pass
            elif token in [14, 39]:
                self.enterOuterAlt(localctx, 2)
                self.state = 175
                self.predicate_statement()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Filter_statementContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return DroltaParser.RULE_filter_statement

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class AndFilterContext(Filter_statementContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DroltaParser.Filter_statementContext
            super().__init__(parser)
            self.left = None # Filter_statementContext
            self.right = None # Filter_statementContext
            self.copyFrom(ctx)

        def OPEN_PAR(self):
            return self.getToken(DroltaParser.OPEN_PAR, 0)
        def AND(self):
            return self.getToken(DroltaParser.AND, 0)
        def CLOSE_PAR(self):
            return self.getToken(DroltaParser.CLOSE_PAR, 0)
        def filter_statement(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DroltaParser.Filter_statementContext)
            else:
                return self.getTypedRuleContext(DroltaParser.Filter_statementContext,i)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAndFilter" ):
                listener.enterAndFilter(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAndFilter" ):
                listener.exitAndFilter(self)


    class ComparisonFilterContext(Filter_statementContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DroltaParser.Filter_statementContext
            super().__init__(parser)
            self.left = None # VariableContext
            self.op = None # Comparison_operatorContext
            self.right = None # AtomContext
            self.copyFrom(ctx)

        def OPEN_PAR(self):
            return self.getToken(DroltaParser.OPEN_PAR, 0)
        def CLOSE_PAR(self):
            return self.getToken(DroltaParser.CLOSE_PAR, 0)
        def variable(self):
            return self.getTypedRuleContext(DroltaParser.VariableContext,0)

        def comparison_operator(self):
            return self.getTypedRuleContext(DroltaParser.Comparison_operatorContext,0)

        def atom(self):
            return self.getTypedRuleContext(DroltaParser.AtomContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterComparisonFilter" ):
                listener.enterComparisonFilter(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitComparisonFilter" ):
                listener.exitComparisonFilter(self)


    class NotFilterContext(Filter_statementContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DroltaParser.Filter_statementContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def OPEN_PAR(self):
            return self.getToken(DroltaParser.OPEN_PAR, 0)
        def NOT(self):
            return self.getToken(DroltaParser.NOT, 0)
        def filter_statement(self):
            return self.getTypedRuleContext(DroltaParser.Filter_statementContext,0)

        def CLOSE_PAR(self):
            return self.getToken(DroltaParser.CLOSE_PAR, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterNotFilter" ):
                listener.enterNotFilter(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitNotFilter" ):
                listener.exitNotFilter(self)


    class OrFilterContext(Filter_statementContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DroltaParser.Filter_statementContext
            super().__init__(parser)
            self.left = None # Filter_statementContext
            self.right = None # Filter_statementContext
            self.copyFrom(ctx)

        def OPEN_PAR(self):
            return self.getToken(DroltaParser.OPEN_PAR, 0)
        def OR(self):
            return self.getToken(DroltaParser.OR, 0)
        def CLOSE_PAR(self):
            return self.getToken(DroltaParser.CLOSE_PAR, 0)
        def filter_statement(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DroltaParser.Filter_statementContext)
            else:
                return self.getTypedRuleContext(DroltaParser.Filter_statementContext,i)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterOrFilter" ):
                listener.enterOrFilter(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitOrFilter" ):
                listener.exitOrFilter(self)


    class InFilterContext(Filter_statementContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DroltaParser.Filter_statementContext
            super().__init__(parser)
            self.left = None # VariableContext
            self.copyFrom(ctx)

        def OPEN_PAR(self):
            return self.getToken(DroltaParser.OPEN_PAR, 0)
        def IN(self):
            return self.getToken(DroltaParser.IN, 0)
        def atom_list(self):
            return self.getTypedRuleContext(DroltaParser.Atom_listContext,0)

        def CLOSE_PAR(self):
            return self.getToken(DroltaParser.CLOSE_PAR, 0)
        def variable(self):
            return self.getTypedRuleContext(DroltaParser.VariableContext,0)

        def NOT(self):
            return self.getToken(DroltaParser.NOT, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterInFilter" ):
                listener.enterInFilter(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitInFilter" ):
                listener.exitInFilter(self)



    def filter_statement(self):

        localctx = DroltaParser.Filter_statementContext(self, self._ctx, self.state)
        self.enterRule(localctx, 28, self.RULE_filter_statement)
        self._la = 0 # Token type
        try:
            self.state = 210
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,26,self._ctx)
            if la_ == 1:
                localctx = DroltaParser.ComparisonFilterContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 178
                self.match(DroltaParser.OPEN_PAR)
                self.state = 179
                localctx.left = self.variable()
                self.state = 180
                localctx.op = self.comparison_operator()
                self.state = 181
                localctx.right = self.atom()
                self.state = 182
                self.match(DroltaParser.CLOSE_PAR)
                pass

            elif la_ == 2:
                localctx = DroltaParser.InFilterContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 184
                self.match(DroltaParser.OPEN_PAR)
                self.state = 185
                localctx.left = self.variable()
                self.state = 187
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==14:
                    self.state = 186
                    self.match(DroltaParser.NOT)


                self.state = 189
                self.match(DroltaParser.IN)
                self.state = 190
                self.atom_list()
                self.state = 191
                self.match(DroltaParser.CLOSE_PAR)
                pass

            elif la_ == 3:
                localctx = DroltaParser.AndFilterContext(self, localctx)
                self.enterOuterAlt(localctx, 3)
                self.state = 193
                self.match(DroltaParser.OPEN_PAR)
                self.state = 194
                localctx.left = self.filter_statement()
                self.state = 195
                self.match(DroltaParser.AND)
                self.state = 196
                localctx.right = self.filter_statement()
                self.state = 197
                self.match(DroltaParser.CLOSE_PAR)
                pass

            elif la_ == 4:
                localctx = DroltaParser.OrFilterContext(self, localctx)
                self.enterOuterAlt(localctx, 4)
                self.state = 199
                self.match(DroltaParser.OPEN_PAR)
                self.state = 200
                localctx.left = self.filter_statement()
                self.state = 201
                self.match(DroltaParser.OR)
                self.state = 202
                localctx.right = self.filter_statement()
                self.state = 203
                self.match(DroltaParser.CLOSE_PAR)
                pass

            elif la_ == 5:
                localctx = DroltaParser.NotFilterContext(self, localctx)
                self.enterOuterAlt(localctx, 5)
                self.state = 205
                self.match(DroltaParser.OPEN_PAR)
                self.state = 206
                self.match(DroltaParser.NOT)
                self.state = 207
                self.filter_statement()
                self.state = 208
                self.match(DroltaParser.CLOSE_PAR)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Atom_listContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def BRACKET_L(self):
            return self.getToken(DroltaParser.BRACKET_L, 0)

        def BRACKET_R(self):
            return self.getToken(DroltaParser.BRACKET_R, 0)

        def atom(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DroltaParser.AtomContext)
            else:
                return self.getTypedRuleContext(DroltaParser.AtomContext,i)


        def COMMA(self, i:int=None):
            if i is None:
                return self.getTokens(DroltaParser.COMMA)
            else:
                return self.getToken(DroltaParser.COMMA, i)

        def getRuleIndex(self):
            return DroltaParser.RULE_atom_list

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAtom_list" ):
                listener.enterAtom_list(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAtom_list" ):
                listener.exitAtom_list(self)




    def atom_list(self):

        localctx = DroltaParser.Atom_listContext(self, self._ctx, self.state)
        self.enterRule(localctx, 30, self.RULE_atom_list)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 212
            self.match(DroltaParser.BRACKET_L)
            self.state = 221
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la) & ~0x3f) == 0 and ((1 << _la) & 8177617731586) != 0):
                self.state = 213
                self.atom()
                self.state = 218
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==9:
                    self.state = 214
                    self.match(DroltaParser.COMMA)
                    self.state = 215
                    self.atom()
                    self.state = 220
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)



            self.state = 223
            self.match(DroltaParser.BRACKET_R)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Comparison_operatorContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def GT(self):
            return self.getToken(DroltaParser.GT, 0)

        def GTE(self):
            return self.getToken(DroltaParser.GTE, 0)

        def LT(self):
            return self.getToken(DroltaParser.LT, 0)

        def LTE(self):
            return self.getToken(DroltaParser.LTE, 0)

        def EQ(self):
            return self.getToken(DroltaParser.EQ, 0)

        def NEQ(self):
            return self.getToken(DroltaParser.NEQ, 0)

        def getRuleIndex(self):
            return DroltaParser.RULE_comparison_operator

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterComparison_operator" ):
                listener.enterComparison_operator(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitComparison_operator" ):
                listener.exitComparison_operator(self)




    def comparison_operator(self):

        localctx = DroltaParser.Comparison_operatorContext(self, self._ctx, self.state)
        self.enterRule(localctx, 32, self.RULE_comparison_operator)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 225
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & 1056964608) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Predicate_statementContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return DroltaParser.RULE_predicate_statement

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class PredicateNotContext(Predicate_statementContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DroltaParser.Predicate_statementContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def NOT(self):
            return self.getToken(DroltaParser.NOT, 0)
        def predicate_statement(self):
            return self.getTypedRuleContext(DroltaParser.Predicate_statementContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPredicateNot" ):
                listener.enterPredicateNot(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPredicateNot" ):
                listener.exitPredicateNot(self)


    class PredicateContext(Predicate_statementContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a DroltaParser.Predicate_statementContext
            super().__init__(parser)
            self.PredicateName = None # Token
            self.copyFrom(ctx)

        def OPEN_PAR(self):
            return self.getToken(DroltaParser.OPEN_PAR, 0)
        def CLOSE_PAR(self):
            return self.getToken(DroltaParser.CLOSE_PAR, 0)
        def IDENTIFIER(self):
            return self.getToken(DroltaParser.IDENTIFIER, 0)
        def predicate_param(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DroltaParser.Predicate_paramContext)
            else:
                return self.getTypedRuleContext(DroltaParser.Predicate_paramContext,i)

        def COMMA(self, i:int=None):
            if i is None:
                return self.getTokens(DroltaParser.COMMA)
            else:
                return self.getToken(DroltaParser.COMMA, i)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPredicate" ):
                listener.enterPredicate(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPredicate" ):
                listener.exitPredicate(self)



    def predicate_statement(self):

        localctx = DroltaParser.Predicate_statementContext(self, self._ctx, self.state)
        self.enterRule(localctx, 34, self.RULE_predicate_statement)
        self._la = 0 # Token type
        try:
            self.state = 242
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [39]:
                localctx = DroltaParser.PredicateContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 227
                localctx.PredicateName = self.match(DroltaParser.IDENTIFIER)
                self.state = 228
                self.match(DroltaParser.OPEN_PAR)
                self.state = 237
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==39:
                    self.state = 229
                    self.predicate_param()
                    self.state = 234
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    while _la==9:
                        self.state = 230
                        self.match(DroltaParser.COMMA)
                        self.state = 231
                        self.predicate_param()
                        self.state = 236
                        self._errHandler.sync(self)
                        _la = self._input.LA(1)



                self.state = 239
                self.match(DroltaParser.CLOSE_PAR)
                pass
            elif token in [14]:
                localctx = DroltaParser.PredicateNotContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 240
                self.match(DroltaParser.NOT)
                self.state = 241
                self.predicate_statement()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Predicate_paramContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self):
            return self.getToken(DroltaParser.IDENTIFIER, 0)

        def EQ(self):
            return self.getToken(DroltaParser.EQ, 0)

        def atom(self):
            return self.getTypedRuleContext(DroltaParser.AtomContext,0)


        def getRuleIndex(self):
            return DroltaParser.RULE_predicate_param

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPredicate_param" ):
                listener.enterPredicate_param(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPredicate_param" ):
                listener.exitPredicate_param(self)




    def predicate_param(self):

        localctx = DroltaParser.Predicate_paramContext(self, self._ctx, self.state)
        self.enterRule(localctx, 36, self.RULE_predicate_param)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 244
            self.match(DroltaParser.IDENTIFIER)
            self.state = 245
            self.match(DroltaParser.EQ)
            self.state = 246
            self.atom()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class AtomContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def variable(self):
            return self.getTypedRuleContext(DroltaParser.VariableContext,0)


        def INT_LITERAL(self):
            return self.getToken(DroltaParser.INT_LITERAL, 0)

        def FLOAT_LITERAL(self):
            return self.getToken(DroltaParser.FLOAT_LITERAL, 0)

        def STRING_LITERAL(self):
            return self.getToken(DroltaParser.STRING_LITERAL, 0)

        def NULL(self):
            return self.getToken(DroltaParser.NULL, 0)

        def TRUE(self):
            return self.getToken(DroltaParser.TRUE, 0)

        def FALSE(self):
            return self.getToken(DroltaParser.FALSE, 0)

        def getRuleIndex(self):
            return DroltaParser.RULE_atom

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAtom" ):
                listener.enterAtom(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAtom" ):
                listener.exitAtom(self)




    def atom(self):

        localctx = DroltaParser.AtomContext(self, self._ctx, self.state)
        self.enterRule(localctx, 38, self.RULE_atom)
        try:
            self.state = 255
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [1]:
                self.enterOuterAlt(localctx, 1)
                self.state = 248
                self.variable()
                pass
            elif token in [40]:
                self.enterOuterAlt(localctx, 2)
                self.state = 249
                self.match(DroltaParser.INT_LITERAL)
                pass
            elif token in [41]:
                self.enterOuterAlt(localctx, 3)
                self.state = 250
                self.match(DroltaParser.FLOAT_LITERAL)
                pass
            elif token in [42]:
                self.enterOuterAlt(localctx, 4)
                self.state = 251
                self.match(DroltaParser.STRING_LITERAL)
                pass
            elif token in [36]:
                self.enterOuterAlt(localctx, 5)
                self.state = 252
                self.match(DroltaParser.NULL)
                pass
            elif token in [37]:
                self.enterOuterAlt(localctx, 6)
                self.state = 253
                self.match(DroltaParser.TRUE)
                pass
            elif token in [38]:
                self.enterOuterAlt(localctx, 7)
                self.state = 254
                self.match(DroltaParser.FALSE)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class VariableContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self):
            return self.getToken(DroltaParser.IDENTIFIER, 0)

        def getRuleIndex(self):
            return DroltaParser.RULE_variable

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterVariable" ):
                listener.enterVariable(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitVariable" ):
                listener.exitVariable(self)




    def variable(self):

        localctx = DroltaParser.VariableContext(self, self._ctx, self.state)
        self.enterRule(localctx, 40, self.RULE_variable)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 257
            self.match(DroltaParser.T__0)
            self.state = 258
            self.match(DroltaParser.IDENTIFIER)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








"""Drolta: Getting Started

This script contains a sample to get users started with using Drolta. It uses
characters from HBO's House of the Dragon to fill a SQLite database with test
data that is similar to what someone might find in a simulationist emergent
narrative game.

"""

import logging
import sqlite3

import drolta.engine


def initialize_samples_data(db: sqlite3.Connection) -> None:
    """Initializes the database connection with sample data."""

    cursor = db.cursor()

    cursor.executescript(
        """
        DROP TABLE IF EXISTS characters;
        DROP TABLE IF EXISTS houses;
        DROP TABLE IF EXISTS relations;

        CREATE TABLE characters (
            id INTEGER PRIMARY KEY NOT NULL,
            name TEXT,
            house_id INTEGER,
            sex TEXT,
            life_stage TEXT,
            is_alive INTEGER,
            FOREIGN KEY (house_id) REFERENCES houses(id)
        ) STRICT;

        CREATE TABLE houses (
            id INTEGER NOT NULL PRIMARY KEY,
            name TEXT NOT NULL,
            reputation INT NOT NULL,
            is_noble INT NOT NULL
        ) STRICT;

        CREATE TABLE relations (
            from_id INTEGER NOT NULL,
            to_id INTEGER NOT NULL,
            type TEXT NOT NULL,
            FOREIGN KEY (from_id) REFERENCES characters(id),
            FOREIGN KEY (to_id) REFERENCES characters(id)
        ) STRICT;
        """
    )

    cursor.executemany(
        """
        INSERT INTO
        characters (id, name, house_id, sex, life_stage, is_alive)
        VALUES
        (?, ?, ?, ?, ?, ?);
        """,
        [
            (1, "Rhaenyra", 1, "F", "Adult", 1),
            (2, "Laenor", 2, "M", "Adult", 1),
            (3, "Harwin", 3, "M", "Adult", 1),
            (4, "Jacaerys", 2, "M", "Teen", 1),
            (5, "Addam", None, "M", "Teen", 1),
            (6, "Corlys", 2, "M", "Adult", 1),
            (7, "Marilda", None, "F", "Adult", 0),
            (8, "Alyn", None, "M", "Adult", 1),
            (9, "Rhaenys", 1, "F", "Adult", 0),
            (10, "Laena", 2, "F", "Adult", 0),
            (11, "Daemon", 1, "M", "Adult", 1),
            (12, "Baela", 1, "F", "Teen", 1),
            (13, "Viserys", 1, "M", "Senior", 0),
            (14, "Alicent", 5, "F", "Adult", 1),
            (15, "Otto", 5, "M", "Senior", 1),
            (16, "Aegon", 1, "M", "Teen", 1),
            (17, "Cristen", 4, "M", "Adult", 1),
        ],
    )

    cursor.executemany(
        """
        INSERT INTO
            houses(id, name, reputation, is_noble)
        VALUES
            (?, ?, ?, ?);
        """,
        [
            (1, "Targaryen", 50, 1),
            (2, "Velaryon", 50, 1),
            (3, "Strong", 50, 0),
            (4, "Cole", 50, 0),
            (5, "Hightower", 50, 1),
            (6, "Belmont", 50, 1),
        ],
    )

    cursor.executemany(
        """
        INSERT INTO
            relations (from_id, to_id, type)
        VALUES
            (?, ?, ?);
        """,
        [
            (4, 1, "Mother"),  # Jace -> Rhaenyra
            (4, 2, "Father"),  # Jace -> Laenor
            (4, 3, "BiologicalFather"),  # Jace -> Harwin
            (5, 6, "BiologicalFather"),  # Addam -> Corlys
            (2, 6, "BiologicalFather"),  # Laenor -> Corlys
            (2, 6, "Father"),  # Laenor -> Corlys
            (5, 7, "Mother"),  # Addam -> Marilda
            (8, 7, "Mother"),  # Alyn -> Marilda
            (8, 6, "BiologicalFather"),  # Alyn -> Corlys
            (2, 9, "Mother"),  # Laenor -> Rhaenys
            (10, 9, "Mother"),  # Laena -> Rhaenys
            (10, 6, "Father"),  # Laena -> Corlys
            (10, 6, "BiologicalFather"),  # Laena -> Corlys
            (6, 9, "Widower"),  # Corlys -> Rhaenys
            (6, 9, "FormerSpouse"),  # Corlys -> Rhaenys
            (9, 6, "FormerSpouse"),  # Rhaenys -> Corlys
            (12, 10, "Mother"),  # Baela -> Laena
            (12, 11, "Father"),  # Baela -> Daemon
            (12, 11, "BiologicalFather"),  # Baela -> Daemon
            (1, 11, "Spouse"),  # Rhaenyra -> Daemon
            (11, 1, "Spouse"),  # Daemon -> Rhaenyra
            (10, 11, "FormerSpouse"),  # Laena -> Daemon
            (11, 10, "FormerSpouse"),  # Daemon -> Laena
            (11, 10, "Widower"),  # Daemon -> Laena
            (1, 13, "Father"),  # Rhaenyra -> Viserys
            (1, 13, "BiologicalFather"),  # Rhaenyra -> Viserys
            (16, 14, "Mother"),  # Aegon -> Alicent
            (14, 15, "Father"),  # Alicent => Otto
            (14, 15, "BiologicalFather"),  # Alicent => Otto
        ],
    )

    db.commit()


def main() -> None:
    """Main Function."""

    logging.basicConfig(level=logging.INFO)

    # First, create a new SQLite database connection.
    # The database doesn't need to be in-memory. We use
    # an in-memory database for simplicity.
    db = sqlite3.Connection(":memory:")

    initialize_samples_data(db)

    engine = drolta.engine.QueryEngine()

    engine.execute_script(
        """
        ALIAS characters AS Character;
        ALIAS relations AS Relation;
        ALIAS houses AS House;

        DEFINE
            PaternalHalfSiblings(?x, ?y)
        WHERE
            Relation(from_id=?x, to_id=?bf, type="BiologicalFather")
            Relation(from_id=?y, to_id=?bf, type="BiologicalFather")
            Relation(from_id=?x, to_id=?x_m, type="Mother")
            Relation(from_id=?y, to_id=?y_m, type="Mother")
            ((?x_m != ?y_m) AND (?x != ?y));

        DEFINE
            FamilySize(?family_id AS id, COUNT(?character_id) AS count)
        WHERE
            Family(id=?family_id)
            Character(id=?character_id, family_id=?family_id)
        """
    )

    # Query the database for all paternal half-siblings of the character
    # named "Addam". This is done by using the rule we specified above
    # and using the AND operator to ensure that the character has the name
    # Addam.
    with engine.query(
        """
        FIND
        ?siblingId, ?siblingName
        WHERE
            Character(id=?adam_id, name="Addam")
            PaternalHalfSiblings(x=?adam_id, y=?siblingId)
            Character(id=?siblingId, name=?siblingName)
        ORDER BY ?siblingId;
        """,
        db,
    ) as result:

        # Get the actual rows in the result. The order of the columns is the
        # same as the order of the variables specified after FIND. In this case,
        # the first column is the character's ID, and the second is their name.
        print("Addam's Half Siblings:")
        for sibling_id, sibling_name in result.fetch_all():
            print(f"ID: {sibling_id}, Name: {sibling_name}")

    # Output:
    #
    # ID: 2, Name: Laenor
    # ID: 10, Name: Laena

    db.close()


if __name__ == "__main__":
    main()



"""Drolta SQLite Database Wrappers.

The way that Drolta wraps database instances is adapted from Pandas.

References:
- https://github.com/pandas-dev/pandas/blob/main/pandas/io/sql.py
"""

import sqlite3
from contextlib import contextmanager
from typing import Any, Optional


class SQLiteDatabase:
    """Manages a sqlite database connection."""

    __slots__ = ("conn",)

    conn: sqlite3.Connection
    """The connection to the database."""

    def __init__(self, conn: sqlite3.Connection) -> None:
        super().__init__()
        self.conn = conn

    @contextmanager
    def run_transaction(self):
        """Execute a transaction on the database."""

        cur = self.conn.cursor()
        try:
            yield cur
            self.conn.commit()
        except Exception:
            self.conn.rollback()
            raise
        finally:
            cur.close()

    def execute(self, sql: str, params: Optional[tuple[Any, ...]] = None) -> None:
        """Execute an operation on the database.

        Parameters
        ----------
        sql: str
            A SQL expression.
        params: tuple of Any, optional
            Parameters to pass to the SQL expression.
        """
        with self.run_transaction() as cur:
            if params:
                cur.execute(sql, params)
            else:
                cur.execute(sql)



"""Drolta Interpreter."""

from __future__ import annotations

import logging
import sqlite3
from typing import Any, Generator, Iterable, Optional, cast

import attrs
import sqlparse

from drolta.ast import (
    ASTVisitor,
    DeclareAliasExpression,
    DeclareRuleExpression,
    ExpressionNode,
    ExpressionType,
    NotPredicateExpression,
    PredicateExpression,
    ProgramNode,
    QueryExpression,
    is_filter_expression,
)
from drolta.data import EngineData, ResultVariable, RuleData
from drolta.db import SQLiteDatabase
from drolta.errors import ProgrammingError

_logger = logging.getLogger(__name__)


def get_execution_order(node: ExpressionNode) -> int:
    """Get the order number of the expression (Lower is higher priority)."""
    expression_type = node.get_expression_type()

    if expression_type == ExpressionType.PREDICATE_CALL:
        return 0

    if is_filter_expression(node):
        return 1

    if expression_type == ExpressionType.PREDICATE_NOT:
        return 1

    return 2


def cycle_check_dfs(
    aliases: dict[str, str], value: str, visited: set[str], stack: list[str]
) -> tuple[bool, str]:
    """Perform Depth-First Search to check for cycles."""

    if value not in visited:
        visited.add(value)
        stack.append(value)

        if value in aliases:
            target = aliases[value]

            if target not in visited:
                result = cycle_check_dfs(aliases, target, visited, stack)
                if result[0] is True:
                    return result

            if target in stack:
                return True, target

    stack.pop()
    return False, ""


def has_alias_cycle(aliases: dict[str, str]) -> tuple[bool, str]:
    """Check if the alias dictionary has a cycle."""

    stack: list[str] = []
    visited: set[str] = set()

    for alias in aliases.keys():
        if alias not in visited:
            result = cycle_check_dfs(aliases, alias, visited, stack)
            if result[0] is True:
                return result

    return False, ""


def sqlite_dtype_to_py(d_type: str) -> str:
    """Convert SQLite data type name to python type name."""

    if d_type == "INT":
        return "int"
    elif d_type == "TEXT":
        return "str"
    elif d_type == "REAL":
        return "float"

    return "object"


@attrs.define(frozen=True, slots=True)
class TempResult:
    """Information about an intermediate result of a query."""

    table_name: str
    """The name of the temporary table with this result's data"""
    output_vars: set[str]
    """The column names of the temp table"""

    @staticmethod
    def get_common_vars(b: TempResult, a: TempResult) -> list[str]:
        """Get common variables between two temp results."""
        return sorted(a.output_vars.intersection(b.output_vars))


@attrs.define(frozen=True, slots=True)
class ColumnInfo:
    """Information about a column in a table."""

    name: str
    """The name of the column."""
    d_type: str
    """The data type (int | str)."""


class DroltaResult:
    """The result of a Drolta Query.

    This class wraps a sqlite3 Cursor object to ensure that all temporary
    tables are removed at the start of a new query.
    """

    __slots__ = ("_column_info", "_has_read_data", "_cursor", "_db", "_result_table")

    _column_info: list[ColumnInfo]
    _has_read_data: bool
    _result_table: TempResult
    _db: SQLiteDatabase
    _cursor: Optional[sqlite3.Cursor]

    def __init__(
        self,
        columns: list[ColumnInfo],
        db: SQLiteDatabase,
        result_table: TempResult,
        cursor: Optional[sqlite3.Cursor] = None,
    ) -> None:
        self._column_info = [*columns]
        self._has_read_data = False
        self._cursor = cursor
        self._db = db
        self._result_table = result_table

    @property
    def description(self) -> Iterable[ColumnInfo]:
        """Get information about the columns in the result."""
        return [*self._column_info]

    def fetch_all(self) -> list[Any]:
        """Get all results from the last query."""

        if self._has_read_data:
            raise RuntimeError("Data already fetched from this result.")

        self._has_read_data = True

        if self._cursor is None:
            return []

        result = self._cursor.fetchall()

        self.destroy()

        return result

    def fetch_chunks(self, size: int) -> Generator[list[Any], Any, None]:
        """Fetch the next batch of results."""

        if self._has_read_data:
            raise RuntimeError("Data already fetched from this result.")

        self._has_read_data = True

        if self._cursor is None:
            yield []
            return

        next_batch = self._cursor.fetchmany(size)

        while next_batch:
            yield next_batch
            next_batch = self._cursor.fetchmany(size)

        self.destroy()

    def destroy(self):
        """Destroy this result, freeing resources."""
        if self._cursor:
            self._cursor.close()
            self._db.execute(f"DROP TABLE IF EXISTS {self._result_table.table_name};")
            self._cursor = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any):
        self.destroy()


class FormattedSqlString:
    """Represents a SQL string."""

    __slots__ = ("raw_sql",)

    raw_sql: str

    def __init__(self, raw_sql: str) -> None:
        self.raw_sql = raw_sql

    def __str__(self) -> str:
        return sqlparse.format(self.raw_sql, reindent=True, keyword_case="upper")  # type: ignore


@attrs.define(slots=True)
class Scope:
    """Information about the current variable scope of the query."""

    scope_id: int
    """The ID of the current scope."""
    output_vars: list[ResultVariable] = attrs.field(factory=list)
    """Variables output by this scope."""
    tables: list[TempResult] = attrs.field(factory=list)
    """Temporary result tables."""
    next_table_id: int = 1
    """The ID assigned to the next table in this scope."""


SUPPORTED_AGGREGATES = ("COUNT", "MAX", "MIN", "AVG", "SUM")
"""Aggregate functions supported by Drolta."""


class ScriptInterpreter(ASTVisitor):
    """Interpreter used for scripts."""

    __slots__ = ("engine_data",)

    engine_data: EngineData
    """State data for the query engine."""

    def __init__(self, engine_data: EngineData) -> None:
        super().__init__()
        self.engine_data = engine_data

    def visit_program(self, node: ProgramNode):
        for child in node.children:
            self.visit(child)

    def visit_declare_alias(self, node: DeclareAliasExpression):
        new_alias_dict = {**self.engine_data.aliases, node.alias: node.original_name}

        has_cycle, cycled_alias = has_alias_cycle(new_alias_dict)

        if has_cycle:
            raise ProgrammingError(f"Circular aliases for: {cycled_alias}.")

        self.engine_data.aliases[node.alias] = node.original_name

        _logger.debug("Declared alias: %s -> %s", node.alias, node.original_name)

    def visit_declare_rule(self, node: DeclareRuleExpression):
        result_vars: list[ResultVariable] = []
        for entry in node.result_vars:
            result_vars.append(
                ResultVariable(
                    var_name=entry.var_name,
                    aggregate_name=entry.aggregate_name,
                    alias=entry.alias,
                )
            )

        rule = RuleData(
            name=node.name,
            result_vars=result_vars,
            where_expressions=[*node.where_expressions],
            group_by=node.group_by,
            order_by=node.order_by,
            limit=node.limit,
        )

        self.engine_data.rules[rule.name] = rule

        _logger.debug("Declared rule: %s", rule.name)

    def visit_query(self, node: QueryExpression):
        raise ProgrammingError("Queries not allowed while executing scripts.")


_next_query_id: int = 0
"""Gives each query a unique ID to prevent table name clashes."""


class QueryInterpreter(ASTVisitor):
    """Interpreter used for queries."""

    TEMP_TABLE_PREFIX = "temp__"
    """Name prefix for all temporary tables created by the query engine."""

    __slots__ = ("db", "engine_data", "scope_stack", "result", "bindings")

    db: SQLiteDatabase
    """Database connection."""
    engine_data: EngineData
    """State data for the query engine."""
    scope_stack: list[Scope]
    """Stack of scopes used during query evaluation."""
    result: DroltaResult
    """Cursor with results."""
    bindings: dict[str, Any]
    """Variable bindings supplied by the user."""

    def __init__(
        self, db: SQLiteDatabase, engine_data: EngineData, bindings: dict[str, Any]
    ) -> None:
        super().__init__()
        self.db = db
        self.engine_data = engine_data
        self.scope_stack = []
        self.result = DroltaResult([], db, TempResult("", set()))
        self.bindings = bindings

    def visit_declare_alias(self, node: DeclareAliasExpression):
        raise ProgrammingError("Alias declarations not allowed while querying.")

    def visit_declare_rule(self, node: DeclareRuleExpression):
        raise ProgrammingError("Rule declarations not allowed while querying.")

    def visit_query(self, node: QueryExpression):
        global _next_query_id  # pylint: disable=W0603
        _next_query_id += 1
        self.clear_scope_stack()

        _logger.debug("Evaluating query.")

        scope = self.new_scope()

        for entry in node.result_vars:
            scope.output_vars.append(
                ResultVariable(
                    var_name=entry.var_name,
                    aggregate_name=entry.aggregate_name,
                    alias=entry.alias,
                )
            )

        # Sort expressions to push filters and not-predicate expressions
        # to the end of the query.
        where_expressions = sorted(node.where_expressions, key=get_execution_order)

        for expr in where_expressions:
            expression_type = expr.get_expression_type()

            if expression_type == ExpressionType.PREDICATE_CALL:
                self.dispatch_visit_predicate(cast(PredicateExpression, expr))

            elif is_filter_expression(expr):
                self.visit_filter(expr)

            elif expression_type == ExpressionType.PREDICATE_NOT:
                self.visit_not_predicate(cast(NotPredicateExpression, expr))

            self._attempt_join_latest_table()

        self._force_join_all_tables()

        result_table = self.get_scope().tables.pop()

        # For the end of the query, we want to select from the result table
        # all the variables in output vars and assign them to any aliases

        output_cols: list[str] = []
        for entry in self.get_scope().output_vars:
            if entry.aggregate_name:
                if entry.aggregate_name not in SUPPORTED_AGGREGATES:
                    raise ProgrammingError(f"Use of unsupported aggregate: {entry}")

            output_cols.append(str(entry))

        sql_statement = f"""
            SELECT DISTINCT {', '.join(output_cols)}
            FROM {result_table.table_name}
            """

        if self.bindings:
            where_statements: list[str] = []
            for var_name, value in self.bindings.items():
                if value is None:
                    where_statements.append(f"{var_name[1:]} IS NULL")
                elif isinstance(value, str):
                    where_statements.append(f'{var_name[1:]} = "{value}"')
                else:
                    where_statements.append(f"{var_name[1:]} = {value}")

            sql_statement += f"WHERE {' AND '.join(where_statements)}\n"

        if node.group_by:
            sql_statement += f"{node.group_by} "

        if node.order_by:
            sql_statement += f"{node.order_by} "

        if node.limit:
            sql_statement += f"{node.limit} "

        sql_statement += ";"

        _logger.debug(
            "Calculating Final Output:\n{}".format(  # pylint: disable=W1202, C0209
                FormattedSqlString(sql_statement)
            )
        )

        cursor = self.db.conn.cursor()

        result = cursor.execute(sql_statement)

        column_info_cursor = self.db.conn.cursor()

        raw_column_data: dict[str, str] = {
            k: v
            for k, v in column_info_cursor.execute(
                f"SELECT name, type FROM pragma_table_info('{result_table.table_name}')"
            ).fetchall()
        }

        column_info: list[ColumnInfo] = []
        for entry in self.get_scope().output_vars:
            if entry.var_name in raw_column_data:
                if entry.alias:
                    column_info.append(
                        ColumnInfo(
                            entry.alias,
                            sqlite_dtype_to_py(raw_column_data[entry.var_name]),
                        )
                    )
                else:
                    column_info.append(
                        ColumnInfo(
                            entry.var_name,
                            sqlite_dtype_to_py(raw_column_data[entry.var_name]),
                        )
                    )

        self.result = DroltaResult(
            columns=column_info, db=self.db, result_table=result_table, cursor=result
        )

    def visit_program(self, node: ProgramNode):
        self.result = DroltaResult([], db=self.db, result_table=TempResult("", set()))

        for child in node.children:
            self.visit(child)

    def visit_rule(self, name: str, node: PredicateExpression):
        """Evaluates a predicate expression as a rule."""

        self.new_scope()

        rule = self.engine_data.rules[name]

        # Sort expressions to push filters and not-predicate expressions
        # to the end of the query.
        where_expressions = sorted(rule.where_expressions, key=get_execution_order)

        for expr in where_expressions:
            expression_type = expr.get_expression_type()

            if expression_type == ExpressionType.PREDICATE_CALL:
                self.dispatch_visit_predicate(cast(PredicateExpression, expr))

            elif is_filter_expression(expr):
                self.visit_filter(expr)

            elif expression_type == ExpressionType.PREDICATE_NOT:
                self._force_join_all_tables()
                self.visit_not_predicate(cast(NotPredicateExpression, expr))

            self._attempt_join_latest_table()

        self._force_join_all_tables()

        self.execute_rule_sql(node, rule)

    def dispatch_visit_predicate(self, node: PredicateExpression):
        """Manages visiting predicates as true predicates or rules."""

        predicate_name = self.get_final_predicate_name(node.name)

        if predicate_name in self.engine_data.rules:
            self.visit_rule(predicate_name, node)
        else:
            self.visit_predicate(predicate_name, node)

    def visit_predicate(self, predicate_table_name: str, node: PredicateExpression):
        """Evaluate predicate expression against a table in the database."""

        self.execute_predicate_sql(predicate_table_name, node)

    def visit_filter(self, node: ExpressionNode):
        """Evaluate predicate expression."""

        # This function forces all the previous tables to join and performs a filter
        # on them.
        self._force_join_all_tables()

        result_table = self.get_scope().tables.pop()

        sql_statement = f"SELECT * FROM {result_table.table_name} WHERE {node}"

        table_name = self.get_temp_table_name()

        new_result_table = TempResult(
            table_name=table_name, output_vars=set(result_table.output_vars)
        )

        sql_temp_table_statement = (
            f"CREATE TEMPORARY TABLE {table_name} AS {sql_statement};"
        )

        _logger.debug(
            "Filtering Output:\n{}".format(  # pylint: disable=W1202, C0209
                FormattedSqlString(sql_temp_table_statement)
            )
        )

        self.db.execute(sql_temp_table_statement)

        # delete the old temp_table
        self.db.execute(f"DROP TABLE IF EXISTS {result_table.table_name};")

        self.get_scope().tables.append(new_result_table)

    def visit_not_predicate(self, node: NotPredicateExpression):
        """Evaluate predicate expression."""

        self.new_scope()

        expression_type = node.expr.get_expression_type()

        if expression_type == ExpressionType.PREDICATE_CALL:
            self.dispatch_visit_predicate(cast(PredicateExpression, node.expr))

        else:
            raise ProgrammingError("Not statement expects a predicate or rule.")

        result_table = self.get_scope().tables.pop()

        self.pop_scope()

        last_table = self.get_scope().tables.pop()

        table_name = self.get_temp_table_name()

        not_join_result = TempResult(
            table_name=table_name, output_vars=set(last_table.output_vars)
        )

        shared_vars = TempResult.get_common_vars(last_table, result_table)

        if shared_vars:
            where_filters = " AND ".join(
                f"({result_table.table_name}.{v} = {last_table.table_name}.{v})"
                for v in shared_vars
            )

            sql_statement = f"""
                SELECT *
                FROM {last_table.table_name}
                WHERE
                  NOT EXISTS (
                    SELECT 1
                    FROM {result_table.table_name}
                    WHERE {where_filters}
                  )
                """

            sql_temp_table_statement = (
                f"CREATE TEMPORARY TABLE {table_name} AS {sql_statement};"
            )

            _logger.debug(
                "Joining Not Join:\n{}".format(  # pylint: disable=W1202, C0209
                    FormattedSqlString(sql_temp_table_statement)
                )
            )

            self.db.execute(sql_temp_table_statement)

            self.db.execute(f"DROP TABLE IF EXISTS {result_table.table_name};")

            self.db.execute(f"DROP TABLE IF EXISTS {last_table.table_name};")

            self.get_scope().tables.append(not_join_result)

    def _force_join_all_tables(self) -> None:
        """Force all tables in the scope to join."""

        current_scope = self.get_scope()

        if len(current_scope.tables) > 1:
            last_table = current_scope.tables[-1]

            # Create a large join under a new temp_table.

            sql_join_statement = f"SELECT * FROM {last_table.table_name} "

            output_vars: set[str] = set(last_table.output_vars)

            for other_table in current_scope.tables[:-1]:
                shared_vars = TempResult.get_common_vars(last_table, other_table)

                output_vars = output_vars.union(other_table.output_vars)

                if shared_vars:
                    join_cols = " AND ".join(
                        f"({other_table.table_name}.{v} = {last_table.table_name}.{v})"
                        for v in shared_vars
                    )

                    sql_join_statement += (
                        f"JOIN {other_table.table_name} ON {join_cols}"
                    )
                else:
                    sql_join_statement += f"CROSS JOIN {other_table.table_name}"

                sql_join_statement += "\n"

            table_name = self.get_temp_table_name()

            sql_temp_table_statement = (
                f"CREATE TEMPORARY TABLE {table_name} AS {sql_join_statement};"
            )

            _logger.debug(
                "Joining All Tables:\n{}".format(  # pylint: disable=W1202, C0209
                    FormattedSqlString(sql_temp_table_statement)
                )
            )

            self.db.execute(sql_temp_table_statement)

            # Remove all tables involved with the join
            for table in current_scope.tables:
                self.db.execute(f"DROP TABLE IF EXISTS {table.table_name};")

            current_scope.tables.clear()

            current_scope.tables.append(
                TempResult(table_name=table_name, output_vars=output_vars)
            )

    def _attempt_join_latest_table(self) -> None:
        """Try to join the latest table with any existing ones."""
        current_scope = self.get_scope()
        if len(current_scope.tables) > 1:
            # Check if the last table has common vars with any
            # of the previous tables
            last_table = current_scope.tables[-1]

            shared_table_indexes: set[int] = set()
            output_vars: set[str] = set(last_table.output_vars)
            have_shared: list[tuple[int, list[str]]] = []

            for table_idx, other_table in enumerate(current_scope.tables[:-1]):
                shared_vars = TempResult.get_common_vars(last_table, other_table)
                if shared_vars:
                    have_shared.append((table_idx, shared_vars))
                    shared_table_indexes.add(table_idx)
                    output_vars = output_vars.union(other_table.output_vars)

            if have_shared:
                # Create a large join under a new temp_table.

                sql_join_statement = f"SELECT * FROM {last_table.table_name} "

                for idx, var_names in have_shared:
                    temp_table = current_scope.tables[idx]

                    join_cols = " AND ".join(
                        f"({temp_table.table_name}.{v} = {last_table.table_name}.{v})"
                        for v in var_names
                    )

                    sql_join_statement += (
                        f"JOIN {temp_table.table_name} ON {join_cols} "
                    )

                table_name = self.get_temp_table_name()

                sql_temp_table_statement = (
                    f"CREATE TEMPORARY TABLE {table_name} AS {sql_join_statement};"
                )

                _logger.debug(
                    "Joining Tables:\n{}".format(  # pylint: disable=W1202, C0209
                        FormattedSqlString(sql_temp_table_statement)
                    )
                )

                self.db.execute(sql_temp_table_statement)

                # Remove all tables involved with the join
                for idx in range(len(current_scope.tables), -1, -1):
                    if (
                        idx == len(current_scope.tables) - 1
                        or idx in shared_table_indexes
                    ):
                        table = current_scope.tables.pop(idx)

                        self.db.execute(f"DROP TABLE IF EXISTS {table.table_name};")

                current_scope.tables.append(
                    TempResult(table_name=table_name, output_vars=output_vars)
                )

    def execute_predicate_sql(self, table_name: str, node: PredicateExpression) -> None:
        """Execute SQL query for a predicate expression."""

        output_vars: set[str] = set()
        column_statements: list[str] = []
        where_statements: list[str] = []

        for column_name, expr in node.params:
            if expr.get_expression_type() == ExpressionType.VARIABLE:
                column_statements.append(f'{column_name} AS "{expr}"')
                output_vars.add(str(expr))
            else:
                if expr.get_expression_type() == ExpressionType.NULL:
                    where_statements.append(f"{column_name} IS {expr}")
                else:
                    where_statements.append(f"{column_name} = {expr}")

        if len(column_statements) == 0:
            raise ProgrammingError(f"Predicate '{node}' expects one output variable.")

        if where_statements:
            select_expr = f"""
                SELECT {', '.join(column_statements)}
                FROM {table_name}
                WHERE {' AND '.join(where_statements)}
                """
        else:
            select_expr = f"SELECT {', '.join(column_statements)} FROM {table_name} "

        temp_table_name = self.get_temp_table_name()

        sql_statement = f"CREATE TEMPORARY TABLE {temp_table_name} AS {select_expr};"

        _logger.debug(
            "Executing predicate select: \n{}".format(  # pylint: disable=W1202, C0209
                FormattedSqlString(sql_statement)
            )
        )

        self.db.execute(sql_statement)

        self.get_scope().tables.append(TempResult(temp_table_name, output_vars))

    def execute_rule_sql(self, node: PredicateExpression, rule: RuleData) -> None:
        """Get the final SQL expression for a rule expression."""

        # The variables output by this rules
        output_vars: set[str] = set()
        # Where statements used in the final SQL query
        where_statements: list[str] = []
        # Column selection statements
        column_statements: list[str] = []
        # Column selection statements used in the CTE
        cte_column_statements: list[str] = []

        for entry in rule.result_vars:
            if entry.aggregate_name:
                if entry.aggregate_name not in SUPPORTED_AGGREGATES:
                    raise ProgrammingError(f"Use of unsupported aggregate: {entry}")

            cte_column_statements.append(str(entry))

        for column_name, expr in node.params:
            # Add input parameters mapped to variables to the set of output vars
            if expr.get_expression_type() == ExpressionType.VARIABLE:
                column_statements.append(f'{column_name} AS "{expr}"')
                output_vars.add(str(expr))

            # If it is not a variable, then this is a column mapped to a constant
            # and must be added to the WHERE section of the final SQL query
            else:
                if expr.get_expression_type() == ExpressionType.NULL:
                    where_statements.append(f"{column_name} IS {expr}")
                else:
                    where_statements.append(f"{column_name} = {expr}")

        result_table = self.get_scope().tables.pop()

        if len(column_statements) == 0:
            raise ProgrammingError(f"Rule '{node}' expects one output variable.")

        if where_statements:

            select_expr = f"""
                WITH {rule.name} AS (
                  SELECT {', '.join(cte_column_statements)}
                  FROM {result_table.table_name}
                  {' ' + str(rule.group_by) + ' ' if rule.group_by else ''}
                  {' ' + str(rule.order_by) + ' ' if rule.order_by else ''}
                  {' ' + str(rule.limit) + ' ' if rule.limit else ''}
                )
                SELECT {', '.join(column_statements)}
                FROM {rule.name}
                WHERE {' AND '.join(where_statements)}
                """

        else:
            select_expr = f"""
                WITH {rule.name} AS (
                  SELECT {', '.join(cte_column_statements)}
                  FROM {result_table.table_name}
                  {' ' + str(rule.group_by) + ' ' if rule.group_by else ''}
                  {' ' + str(rule.order_by) + ' ' if rule.order_by else ''}
                  {' ' + str(rule.limit) + ' ' if rule.limit else ''}
                )
                SELECT {', '.join(column_statements)}
                FROM {rule.name}
                """

        self.pop_scope()

        temp_table_name = self.get_temp_table_name()

        sql_temp_table_statement = (
            f"CREATE TEMPORARY TABLE {temp_table_name} AS {select_expr};"
        )

        _logger.debug(
            "Executing Rule SQL:\n{}".format(  # pylint: disable=W1202, C0209
                FormattedSqlString(sql_temp_table_statement)
            )
        )

        self.db.execute(sql_temp_table_statement)

        self.db.execute(f"DROP TABLE IF EXISTS {result_table.table_name};")

        self.get_scope().tables.append(TempResult(temp_table_name, output_vars))

    def get_final_predicate_name(self, name: str) -> str:
        """Resolve the final name of a predicate from a potential alias."""

        final_name = name

        while final_name in self.engine_data.aliases:
            final_name = self.engine_data.aliases[final_name]

        return final_name

    def get_temp_table_name(self) -> str:
        """Generate a name for the next temporary table."""
        current_scope = self.get_scope()

        table_name = (
            self.TEMP_TABLE_PREFIX
            + f"{_next_query_id}_"
            + f"{current_scope.scope_id}_"
            + str(current_scope.next_table_id)
        )

        current_scope.next_table_id += 1

        return table_name

    def get_scope(self) -> Scope:
        """Get the current scope."""
        if self.scope_stack:
            return self.scope_stack[-1]

        return self.new_scope()

    def new_scope(self) -> Scope:
        """Push a new scope on the stack."""
        if self.scope_stack:
            scope = Scope(scope_id=self.scope_stack[-1].scope_id + 1)
        else:
            scope = Scope(scope_id=0)

        self.scope_stack.append(scope)
        return scope

    def pop_scope(self) -> Scope:
        """Pop the current scope."""
        scope = self.scope_stack.pop()
        self.destroy_temporary_tables(scope)
        return scope

    def clear_scope_stack(self) -> None:
        """Clear all scopes from the stack."""

        while self.scope_stack:
            self.pop_scope()

    def destroy_temporary_tables(self, scope: Scope) -> None:
        """Destroy all temporary tables at the given scope."""

        while scope.tables:
            table = scope.tables.pop()
            self.db.execute(f"DROP TABLE IF EXISTS {table.table_name};")
